function isHexColorTransparent(hex) {
   const alpha = hex.substring(7);
   return alpha ? alpha === "00" : false;
}

function isHexColorOpaque(hex) {
   const alpha = hex.substring(7);
   return alpha ? alpha === "FF" : true;
}

function getColorAlpha(hex) {
   const alpha = hex.substring(7);
   return alpha ? parseInt(alpha, 16) / 256 : 1;
}

function splitColor(full_color) {
   let color = full_color.slice(0, 7); 
   let alpha = parseInt(full_color.slice(7), 16) / 256;
   return [color, isNaN(alpha) ? 1 : alpha];
}

function getIntColor(image_data, x, y) {
  var start = ((y*image_data.width)+x)*4;
  
  return 	0x100000000 + 
        (image_data.data[start+3] << 24) + 
        (image_data.data[start] << 16) +
        (image_data.data[start+1] << 8) +
        (image_data.data[start+2]);
};

function setIntColor4(image_data, x, y, alpha, r, g, b) {
  var start = ((y*image_data.width)+x)*4;
  image_data.data[start] = r;
  image_data.data[start+1] = g;
  image_data.data[start+2] = b;
  image_data.data[start+3] = alpha;
};

function setIntColor3(image_data, x, y, r, g, b) {
  setIntColor4(image_data, x, y, 255, r, g, b);
};

function setIntColor2(image_data, x, y, alpha, color) {
  var r = (color & 0x00FF0000) >> 16;
  var g = (color & 0x0000FF00) >> 8;
  var b = color & 0x000000FF;
  setIntColor4(image_data, x, y, alpha, r, g, b);
};

function setIntColor1(image_data, x, y, color) {
  var a = (color & 0xFF000000) >>> 24;
  var r = (color & 0x00FF0000) >> 16;
  var g = (color & 0x0000FF00) >> 8;
  var b = color & 0x000000FF;
  setIntColor4(image_data, x, y, a, r, g, b);
};

function setIntColor(image_data, x, y, a1, a2, a3, a4) {
  if(a2 == null) {
     setIntColor1(image_data, x, y, a1);
  } else if(a3 == null && a4 == null) {
     setIntColor2(image_data, x, y, a1, a2);
  }
  else if(a4 == null) {
     setIntColor3(image_data, x, y, a1, a2, a3);
  }
  else {
     setIntColor4(image_data, x, y, a1, a2, a3, a4);
  }
};

//region Helper functions
function safe_key_union(sourceA, sourceB) {
   const aBad = sourceA === undefined || sourceA === null;
   const bBad = sourceB === undefined || sourceB === null;

   if (aBad && bBad) return new Set();

   if (aBad) {
       const bKeys = Object.keys(sourceB);
       return new Set(bKeys);
   }
   const aKeys = Object.keys(sourceA);

   if (bBad) {
       return new Set(aKeys);
   }
   const bKeys = Object.keys(sourceB);

   let setA = new Set(aKeys);
   let setB = new Set(bKeys);

   for (let elem of setB) {
       setA.add(elem);
   }
   return setA;
}

function arrIntersects(a, b) {
   return a.some(item => b.includes(item));
}

//endregion

class Mask {

   constructor() {
       this.container = {};
   }

   setRed(x, y) {
       this.ensureAvailable(x, y);
       this.container[x][y] |= 0x00FF0000;
   }

   setGreen(x, y) {
       this.ensureAvailable(x, y);
       this.container[x][y] |= 0x0000FF00;
   }

   setBlue(x, y) {
       this.ensureAvailable(x, y);
       this.container[x][y] |= 0x000000FF;
   }

   clearRed(x, y) {
       this.ensureAvailable(x, y);
       this.container[x][y] &= ~0x00FF0000;
   }

   clearGreen(x, y) {
       this.ensureAvailable(x, y);
       this.container[x][y] &= ~0x0000FF00;
   }

   clearBlue(x, y) {
       this.ensureAvailable(x, y);
       this.container[x][y] &= ~0x000000FF;
   }

   testRed(x, y) {
       return x in this.container && y in this.container[x] && (this.container[x][y] & 0x00FF0000) !== 0;
   }

   testGreen(x, y) {
       return x in this.container && y in this.container[x] && (this.container[x][y] & 0x0000FF00) !== 0;
   }

   testBlue(x, y) {
       return x in this.container && y in this.container[x] && (this.container[x][y] & 0x000000FF) !== 0;
   }

   ensureAvailable(x, y) {
       if (!(x in this.container)) {
           this.container[x] = {};
       }
       if (!(y in this.container[x])) {
           this.container[x][y] = 0xFF000000;
       }
   }

   getMask(x, y) {
       if (!(x in this.container)) return null;
       if (!(y in this.container[x])) return null;
       return this.container[x][y];

   }

   setMask(xKey, yKey, value) {
       this.ensureAvailable(xKey, yKey);
       this.container[xKey][yKey] = value;
   }

   union(that) {
       let result = new Mask();

       let xKeys = safe_key_union(this.container, that.container);
       for (const xKey of xKeys) {
           let yKeys = safe_key_union(this.container[xKey], that.container[xKey]);
           for (const yKey of yKeys) {
               const thisValue = this.getMask(xKey, yKey);
               const thatValue = that.getMask(xKey, yKey);

               let value;
               if (thisValue === null || thisValue === undefined) {
                   value = thatValue;
               } else if (thatValue === null || thatValue === undefined) {
                   value = thisValue;
               } else {
                   value = thisValue | thatValue;
               }
               result.setMask(xKey, yKey, value);
           }
       }

       return result;
   }

   calcWidth() {
       let min_x = Number.MAX_VALUE;
       let max_x = -Number.MAX_VALUE;
       for (let x in this.container) {
           if (!this.container.hasOwnProperty(x)) continue;
           x = parseInt(x);
           if (x < min_x) {
               min_x = x;
           }
           if (x > max_x) {
               max_x = x;
           }
       }
       return max_x - min_x;
   }

   calcHeight() {
       let min_y = Number.MAX_VALUE;
       let max_y = -Number.MAX_VALUE;
       for (let x in this.container) {
           if (!this.container.hasOwnProperty(x)) continue;
           x = parseInt(x);

           for (let y in this.container[x]) {
               if (!this.container[x].hasOwnProperty(y)) continue;
               y = parseInt(y);

               if (y < min_y) {
                   min_y = y;
               }
               if (y > max_y) {
                   max_y = y;
               }
           }
       }
       return max_y - min_y;
   }
}

function getBoundingRect(rects) {
   let min_x = Infinity;
   let min_y = Infinity;
   let max_x = -Infinity;
   let max_y = -Infinity;
 
   // Iterate through each rectangle to find the minimum and maximum coordinates
   rects.forEach(rect => {
     min_x = Math.min(min_x, rect.x);
     min_y = Math.min(min_y, rect.y);
     max_x = Math.max(max_x, rect.x + rect.w);
     max_y = Math.max(max_y, rect.y + rect.h);
   });
 
   return new Rectangle(min_x, min_y, max_x - min_x, max_y - min_y);
}

class Rectangle {
   constructor(x, y, w, h) {
       this.x = x;
       this.y = y;
       this.w = w;
       this.h = h;
   }

   coordinateInside(x, y) {
       return !(x < this.x || y < this.y || x >= this.x + this.w || y >= this.y + this.h);
   }

   area() {
       return this.w * this.h;
   }

   scale(w_scalar, h_scalar, method = "ceil") {
       const width_increase = Math[method](this.w * w_scalar) - this.w;
       const height_increase = Math[method](this.h * h_scalar) - this.h;

       this.x -= Math[method](width_increase / 2.0);
       this.y -= Math[method](height_increase / 2.0);

       this.w += width_increase;
       this.h += height_increase;
   }

   scaled(w_scalar, h_scalar, method = "ceil") {
       const new_rect = this.clone();
       new_rect.scale(w_scalar, h_scalar, method);
       return new_rect;
   }

   stretch(width, height) {
       this.x -= width;
       this.w += 2 * width;

       this.y -= height;
       this.h += 2 * height;
   }

   stretched(width, height) {
       const new_rect = new Rectangle(this.x, this.y, this.w, this.h);
       new_rect.stretch(width, height);
       return new_rect;
   }

   resize(width, height, method = "ceil") {
       this.x -= Math[method](width / 2);
       this.w += width;

       this.y -= Math[method](height / 2);
       this.h += height;
   }

   resized(width, height, method = "ceil") {
       const new_rect = new Rectangle(this.x, this.y, this.w, this.h);
       new_rect.resize(width, height, method);
       return new_rect;
   }


   clone() {
       return new Rectangle(this.x, this.y, this.w, this.h);
   }

   global_cropped(x, y, w, h) {
       const result = this.clone();
       result.x = Math.max(result.x, x);
       result.y = Math.max(result.y, y);

       const x2 = x + w;
       const y2 = y + h;

       const result_x2 = result.x + result.w;
       const result_y2 = result.y + result.h;

       const target_result_x2 = Math.min(x2, result_x2);
       const target_result_y2 = Math.min(y2, result_y2);

       result.w = target_result_x2 - result.x;
       result.h = target_result_y2 - result.y;

       return result;
   }

   shifted(x, y) {
       const result = this.clone();
       result.x += x;
       result.y += y;
       return result;
   }

   get x2() {
       return this.x + this.w;
   }

   get y2() {
       return this.y + this.h;
   }

   get cx() {
       return this.x + this.w / 2;
   }

   get cy() {
       return this.y + this.h / 2;
   }
}

class GlitchConfig {
   constructor() {
       this.r_channel_intensity = 0;
       this.g_channel_intensity = 0;
       this.b_channel_intensity = 0;
       this.r_negate_intensity = 0;
       this.g_negate_intensity = 0;
       this.b_negate_intensity = 0;
       this.shift_horizontal = false;
       this.shift_vertical = false;
       this.h_bar_size_range = [0, 0];
       this.v_bar_size_range = [0, 0];
       this.shift_intensity_range = [0, 0];
   }

   set_channel_intensities(value) {
       this.r_channel_intensity = value;
       this.g_channel_intensity = value;
       this.b_channel_intensity = value;
   }

   set_negation_intensities(value) {
       this.r_negate_intensity = value;
       this.g_negate_intensity = value;
       this.b_negate_intensity = value;
   }

   set_all_intensities(value) {
       this.set_channel_intensities(value);
       this.set_negation_intensities(value);
   }
}

class ImageProcessor {
   sobelAreas = this.sobelAreasPrewitt;

   constructor(out, outctx, configurations) {
       this.effects_applied = 0;
       this.out = out;
       this.outctx = outctx;

       this.natural_width = out.width;
       this.natural_height = out.height;

       this.aux = document.createElement('canvas');
       this.aux.width = this.natural_width;
       this.aux.height = this.natural_height;
       this.auxctx = this.aux.getContext('2d');
       this.fill = document.createElement('canvas');
       this.fill.width = this.natural_width;
       this.fill.height = this.natural_height;
       this.fillctx = this.fill.getContext('2d');

       this.bar_configuration = configurations.bar_configuration;
       this.pixel_configuration = configurations.pixel_configuration;
       this.blur_configuration = configurations.blur_configuration;
       this.glitch_configuration = configurations.glitch_configuration;
       this.triangle_configuration = configurations.triangle_configuration;
       this.box_configuration = configurations.box_configuration;
       this.sticker_configuration = configurations.sticker_configuration;
       this.sticker_collections = configurations.sticker_collections;
       this.sobel_configuration = configurations.sobel_configuration;
       this.splatter_configuration = configurations.splatter_configuration;
       this.mixed_configuration = configurations.mixed_configuration;
       this.reverse_mode_configuration = configurations.reverse_mode_configuration;
       this.lock_configuration = configurations.lock_configuration;
       this.word_wall_configuration = configurations.word_wall_configuration;

       this.per_layer_queues = [];
   }

   splitAmount(amount, n, prefix = []) {
       if (n === 1) {
           if (amount > 0) {
               return [[...prefix, amount]];
           } else {
               return [];
           }
       }

       let splits = [];
       for (let i = 1; i <= amount; i++) {
           let new_splits = this.splitAmount(amount - i, n - 1, [...prefix, i]);
           new_splits.forEach(split => {
               if (!splits.find(s => s.every((item, index) => item === split.sort((a, b) => a - b)[index]))) {
                   splits.push(split);
               }
           });
       }

       return splits;
   }

   combinations(arr, n) {
       let stack = [];
       let results = [];

       for (let i = 0; i < arr.length; i++) {
           stack.push({start: i, data: [arr[i]]});
       }

       while (stack.length > 0) {
           let {start, data} = stack.pop();
   
           if (data.length === n) {
               results.push(data);
           } else {
               for (let i = start + 1; i < arr.length; i++) {
                   stack.push({start: i, data: [...data, arr[i]]});
               }
           }
       }

       return results;
   }

   finishReverse() {
       if (this.word_wall_configuration.enabled) {
           this.outctx.globalCompositeOperation = "source-over";
           this.outctx.drawImage(this.ww, 0, 0);
       }
   }

   finishFill(rect, cache) {
       if (!cache.reverse_mode && this.word_wall_configuration.enabled) {
           this.fillctx.globalCompositeOperation = "source-atop";
           this.fillctx.drawImage(this.ww, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
       }

       this.outctx.drawImage(this.fill, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
   }

   handleLineShapes(results, preset_config, cache, finishFill) {
       const MAX_SCALE_FACTOR = 0.25;
       const SCALE_FACTOR_MAX_PENALTY = 1;
       const TWO_POINTS_PENALTY = 2;
       const POINT_DIST_MAX_PENALTY = 1; // Only applied to lines made of 2 points
       const SINGLE_POINT_PENALTY = 4;

       let border_width = Math.round(Math.max(1, Math.sqrt(this.natural_width * this.natural_height) / 1024) * preset_config.border_width);
       let smoothing = Math.round(Math.sqrt(this.natural_width * this.natural_height) / 1024 * preset_config.feathering);
       let should_draw_border = border_width > 0 && !isHexColorTransparent(preset_config.border_color);

       const envelopingLines = (rects) => {
           const bestEnvelopingLine = (rects) => {
               let { point, slope } = this.bestFitLine(rects.map((rect) => rect.x), rects.map((rect) => rect.y));
   
               let angle = Math.atan(slope);
               let left = Infinity, right = -Infinity, max_width = -Infinity;
               rects.forEach((rect) => {
                   let radial_dist = this.radialDist(rect.w / 2, rect.h / 2, angle);
                   let dist_to_line = this.pointToLineDist(rect.cx, rect.cy, point.x, point.y, slope);
                   if (dist_to_line - radial_dist < left) {
                       left = dist_to_line - radial_dist;
                   } 
                   if (right < dist_to_line + radial_dist) {
                       right = dist_to_line + radial_dist;
                   }
                   max_width = Math.max(max_width, radial_dist * 2);
               });
               point.x += Math.sin(angle) * (right + left) / 2;
               point.y += Math.cos(angle) * (right + left) / 2;
               let width = right - left;
               return { 
                   line: { point, slope, width }, 
                   used_rects: rects, 
                   penalty: (width / max_width - 1) / MAX_SCALE_FACTOR * SCALE_FACTOR_MAX_PENALTY
               };
           }

           const NEnvelopingLines = (i, curr_used_rects, curr_penalty, rects, rect_counts, calculated_instructions) => {
               // TODO: if we knew here the lowest penalty found so far across all branches of the tree, we could optimize by returning null immediately if the current penalty is already worse than the best penalty, since a penalty cannot get better, only worse
   
               let rect_count = rect_counts[i];
               
               let instructions;
               if (calculated_instructions[rect_count] != null) {
                   instructions = calculated_instructions[rect_count];
               } else {
                   if (rect_count === 1) {
                       instructions = rects.map((rect) => ({
                           rect: rect,
                           used_rects: [rect],
                           penalty: SINGLE_POINT_PENALTY,
                       }));
                   } else {
                       // TODO: save on memory by passing a closure to the combinations function instead of getting enourmous amounts of arrays back, tho the JS engine might be smart enough to optimize this
                       let rect_combinations = this.combinations(rects, rect_count);
                       if (2 < rect_count) {
                           instructions = rect_combinations.reduce((out, rects) =>{
                               let produced_line = bestEnvelopingLine(rects);
                               if (produced_line.penalty <= SCALE_FACTOR_MAX_PENALTY) {
                                   out.push(produced_line);
                               }
                               return out;
                           }, []);
                       } else {
                           instructions = rect_combinations.map((rects) => {
                               let y_dist = rects[1].cy - rects[0].cy;
                               let x_dist = rects[1].cx - rects[0].cx;
                               let slope = y_dist / x_dist;
                               let angle = Math.atan(slope);
                               let dist = Math.hypot(y_dist, x_dist);
                               return {
                                   line: {
                                       point: { x: rects[0].cx, y: rects[0].cy },
                                       slope: slope,
                                       width: Math.max(this.radialDist(rects[0].w, rects[0].h, angle), this.radialDist(rects[1].w, rects[1].h, angle))
                                   },
                                   used_rects: rects,
                                   penalty: TWO_POINTS_PENALTY + (dist / Math.hypot(this.natural_width, this.natural_height) * POINT_DIST_MAX_PENALTY)
                               }
                           });
                       }
                   }
                   calculated_instructions[rect_count] = instructions;
               }
   
               instructions = instructions.filter((line) => !arrIntersects(curr_used_rects, line.used_rects));
   
               if (instructions.length <= 0) {
                   return {
                       lines: [], 
                       penalty: Infinity 
                   };
               }
   
               if (i < rect_counts.length - 1) {
                   let min_penalty = Infinity;
                   let best_lns_and_rcs = [];
                   for (let instruction of instructions) {
                       let { lines_and_rects, penalty } = NEnvelopingLines(i + 1, 
                               [...curr_used_rects, ...instruction.used_rects], 
                               curr_penalty + instruction.penalty, 
                               rects, 
                               rect_counts, 
                               calculated_instructions);
                       if (penalty < min_penalty) {
                           min_penalty = penalty;
                           best_lns_and_rcs = [{ line: instruction.line, rect: instruction.rect}, ...lines_and_rects];
                       }
                   }
                   return { lines_and_rects: best_lns_and_rcs, penalty: min_penalty };
               } else {
                   let best_instruction = _.min(instructions, (instruction) => instruction.penalty);
                   return { 
                       lines_and_rects: [{ 
                           line: best_instruction.line,
                           rect: best_instruction.rect
                       }],
                       penalty: (curr_penalty + best_instruction.penalty) / rects.length 
                   };
               }
           }
   
           // NOTE: the loop will always return a value, since in the worse case scenario, it will eventually attempt to make only lines of 1 or 2 rects each, which will always succeed
           let cached_possible_outputs = {};
           for (let section_count = 1; section_count < rects.length; section_count++) {
               let rect_counts_arr = this.splitAmount(rects.length, section_count);
               // NOTE: We iterate from back to front because the most even splits are located last in the array, and such splits are preferable
               for (let rect_counts of rect_counts_arr.reverse()) {
                   let output = NEnvelopingLines(0, [], 0, rects, rect_counts, cached_possible_outputs);
                   if (output.penalty !== Infinity) {
                       return output.lines_and_rects;
                   }
               }
           }
       };

       let min_x = Infinity, min_y = Infinity, max_x = -Infinity, max_y = -Infinity;
       const drawLine = (point, slope, width) => {
           let extra_extension;
           if (0 < smoothing) {
               width += smoothing;
               extra_extension = (width + smoothing * 4.8) / 2;
           } else {
               extra_extension = width / 2;
           }
           // TODO: handle vertical lines 
           let [ start, end ] = this.extendLine(point, slope, -extra_extension, this.natural_width + extra_extension, -extra_extension, this.natural_height + extra_extension);

           lines.push({
               start: start,
               end: end,
               width: width,
               slope: slope
           });

           let feather_growth = smoothing + Math.floor(smoothing * 4.8);
           let x1, x2, y1, y2;
           if (start.x < end.x) {
               x1 = Math.max(0,  start.x);
               x2 = Math.min(this.natural_width, end.x);
           } else {
               x1 = Math.max(0, end.x);
               x2 = Math.min(this.natural_width, start.x);
           }
           if (start.y < end.y) {
               y1 = Math.max(0, start.y);
               y2 = Math.min(this.natural_height, end.y);
           } else {
               y1 = Math.max(0, end.y);
               y2 = Math.min(this.natural_height, start.y);
           }

           if (border_width * 2 <= feather_growth) {
               let feather_width = width + Math.floor(smoothing * 4.8);
               min_x = Math.min(min_x, x1 - feather_width / 2);
               min_y = Math.min(min_y, y1 - feather_width / 2);
               max_x = Math.max(max_x, x2 + feather_width / 2);
               max_y = Math.max(max_y, y2 + feather_width / 2);
           } else {
               min_x = Math.min(min_x, x1 - width / 2 - preset_config.border_width);
               min_y = Math.min(min_y, y1 - width / 2 - preset_config.border_width);
               max_x = Math.max(max_x, x2 + width / 2 + preset_config.border_width);
               max_y = Math.max(max_y, y2 + width / 2 + preset_config.border_width);
           }

           this.fillctx.filter = "blur(" + smoothing + "px)";
           this.fillctx.lineWidth = width;
           this.fillctx.beginPath();
           this.fillctx.moveTo(start.x, start.y);
           this.fillctx.lineTo(end.x, end.y);
           this.fillctx.stroke();
       }

       // Group the results by klass indiferent to covered or exposed status
       let results_per_body_part = Object.values(_.groupBy(results, (res) => res.klass.body_part_index));

       // Get only the rects and scale them while at it
       let rects_per_body_part =  results_per_body_part.map((results) => results.map(res => res.rect.scaled(preset_config.scale, preset_config.scale)));

       // Split the rects into single rects and pairable rects that could be made into lines
       let [single_rects, pairable_rects] = _.partition(rects_per_body_part, (rects) => rects.length <= 1);

       this.fillctx.reset();

       this.fillctx.save();

       let lines = [];
       for (let rects of pairable_rects) {
           let lines_and_rects = envelopingLines(rects);

           for (let { line, rect } of lines_and_rects) {
               if (line == null) {
                   single_rects.push([rect]);
                   continue;
               }

               let { point, slope, width } = line;
               drawLine(point, slope, width);
           }
       }

       for (let [rect] of single_rects) {
           let point = { x: rect.cx, y: rect.cy };

           // Set the slope to imitate that of the closest existing line, or to 0 if there are no lines
           let { slope = 0 } = _.min(lines, (line) => this.pointToLineDist(point.x, point.y, line.start.x, line.start.y, line.slope) - line.width / 2);

           let width = slope === 0 ? rect.h : this.envelopingLineWidth(rect.w, rect.h, -1 / slope);
           drawLine(point, slope, width);
       }

       this.fillctx.restore();

       let bounding_rect = new Rectangle(Math.floor(min_x), Math.floor(min_y), Math.ceil(max_x - min_x), Math.ceil(max_y - min_y));

       finishFill(bounding_rect);

       if (should_draw_border) {
           this.auxctx.reset();

           if (0 < smoothing) {
               this.fillctx.reset();
           }
           for (let line of lines) {
               if (0 < smoothing) {
                   this.fillctx.lineWidth = line.width - smoothing;
                   this.fillctx.beginPath();
                   this.fillctx.moveTo(line.start.x, line.start.y);
                   this.fillctx.lineTo(line.end.x, line.end.y);
                   this.fillctx.stroke();
               }

               this.auxctx.lineWidth = line.width - smoothing + border_width * 2;
               this.auxctx.beginPath();
               this.auxctx.moveTo(line.start.x, line.start.y);
               this.auxctx.lineTo(line.end.x, line.end.y);
               this.auxctx.stroke();
           }

           this.auxctx.globalCompositeOperation = "destination-out";
           this.auxctx.drawImage(this.fill, bounding_rect.x, bounding_rect.y, bounding_rect.w, bounding_rect.h, bounding_rect.x, bounding_rect.y, bounding_rect.w, bounding_rect.h);
           this.auxctx.globalCompositeOperation = "source-atop";
           this.auxctx.drawImage(this.clip, bounding_rect.x, bounding_rect.y, bounding_rect.w, bounding_rect.h, bounding_rect.x, bounding_rect.y, bounding_rect.w, bounding_rect.h);
           this.auxctx.fillStyle = preset_config.border_color;
           this.auxctx.fillRect(bounding_rect.x, bounding_rect.y, bounding_rect.w, bounding_rect.h);

           this.outctx.drawImage(this.aux, bounding_rect.x, bounding_rect.y, bounding_rect.w, bounding_rect.h, bounding_rect.x, bounding_rect.y, bounding_rect.w, bounding_rect.h);
       }
   }

   handleShapes(results, preset_config, cache, prepareFill, finishFill) {
       let border_width = Math.round(Math.max(1, Math.sqrt(this.natural_width * this.natural_height) / 1024) * preset_config.border_width);
       let smoothing = Math.round(Math.sqrt(this.natural_width * this.natural_height) / 1024 * preset_config.feathering);
       let rounding = 0;

       // NOTE: We don't draw borders if the shape is heart since just drawing a bigger shape for the border doesn't work well with hearts
       let should_draw_border = border_width > 0 && !isHexColorTransparent(preset_config.border_color) && preset_config.shape !== 3;

       this.fillctx.reset();

       let min_x = Infinity, min_y = Infinity, max_x = -Infinity, max_y = -Infinity;
       let feather_growth = 0 < smoothing ? smoothing + Math.floor(smoothing * 4.8) : 0;
       for (let { rect } of results) {
           switch (preset_config.shape) {
               case 0: { // Rectangle
                   rounding = Math.round((Math.sqrt(this.natural_width * this.natural_height) / 1024) * preset_config.rounding);
                   rect.scale(preset_config.scale, preset_config.scale);
                   break;
               }
               case 1: { // Circle
                   rect.scale(preset_config.scale, preset_config.scale);
                   let rwh = (rect.w + rect.h) / 2;
                   rect.resize((rwh - rect.w), (rwh - rect.h));
                   break;
               }
               case 2: { // Ellipse
                   rect.scale(preset_config.scale, preset_config.scale);
                   break;
               }
               case 3: { // Heart
                   rect.scale(preset_config.scale, preset_config.scale);
                   let rwh = (rect.w + rect.h) / 2;
                   rect.resize((rwh - rect.w), (rwh - rect.h));
                   break;
               }
           }

           rect = rect.resized(smoothing, smoothing, "round");
           let feathered_rect = rect.clone();

           if (0 < smoothing) {
               feathered_rect.resize(smoothing * 4.8, smoothing * 4.8, "round");
           }

           this.fillctx.save();
           this.drawShape(this.fillctx, rect, preset_config.shape, smoothing, rounding);
           this.fillctx.restore();

           if (prepareFill) {
               this.fillctx.save();
               prepareFill(rect, feathered_rect);
               this.fillctx.restore();
           }

           if (border_width <= feather_growth) {
               min_x = Math.min(min_x, feathered_rect.x);
               min_y = Math.min(min_y, feathered_rect.y);
               max_x = Math.max(max_x, feathered_rect.x + feathered_rect.w);
               max_y = Math.max(max_y, feathered_rect.y + feathered_rect.h);
           } else {
               min_x = Math.min(min_x, rect.x - border_width);
               min_y = Math.min(min_y, rect.y - border_width);
               max_x = Math.max(max_x, rect.x + rect.w + border_width * 2);
               max_y = Math.max(max_y, rect.y + rect.h + border_width * 2);
           }
       }
       let bounding_rect = new Rectangle(min_x, min_y, max_x - min_x, max_y - min_y);

       if (finishFill) {
           finishFill(bounding_rect);
       } else {
           this.finishFill(bounding_rect, cache);
       }

       if (should_draw_border) {
           this.auxctx.reset();

           if (0 < smoothing) {
               this.fillctx.reset();
           }
           for (let { rect } of results) {
               if (0 < smoothing) {
                   this.drawShape(this.fillctx, rect, preset_config.shape, 0, rounding);
               }

               rect.resize(border_width * 2, border_width * 2, "round");
               let border_scale_factor = (rect.w + rect.h) / (rect.w - border_width * 2 + rect.h - border_width * 2);
               this.drawShape(this.auxctx, rect, preset_config.shape, 0, rounding * border_scale_factor**2);

               this.auxctx.globalCompositeOperation = "source-atop";
               this.auxctx.drawImage(this.out, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
               this.auxctx.globalCompositeOperation = "destination-out";
               this.auxctx.drawImage(this.fill, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
           }

           this.outctx.drawImage(this.aux, bounding_rect.x, bounding_rect.y, bounding_rect.w, bounding_rect.h, bounding_rect.x, bounding_rect.y, bounding_rect.w, bounding_rect.h);
           this.auxctx.globalCompositeOperation = "source-atop";
           this.auxctx.fillStyle = preset_config.border_color;
           this.auxctx.fillRect(bounding_rect.x, bounding_rect.y, bounding_rect.w, bounding_rect.h);
           this.outctx.drawImage(this.aux, bounding_rect.x, bounding_rect.y, bounding_rect.w, bounding_rect.h, bounding_rect.x, bounding_rect.y, bounding_rect.w, bounding_rect.h);
       }
   }

   drawShape(ctx, rect, shape, smoothing, rounding) {
       ctx.fillStyle = "#000";
       ctx.globalCompositeOperation = "source-over";
       ctx.filter = "blur(" + smoothing + "px)";
       ctx.beginPath();
       switch (shape) {
           case 0: { // Rectangle
               ctx.roundRect(rect.x, rect.y, rect.w, rect.h, rounding);
               break;
           }
           case 1: { // Circle
               ctx.arc(rect.x + rect.w / 2, rect.y + rect.h / 2, rect.w / 2, 0, 2 * Math.PI);
               break;
           }
           case 2: { // Ellipse
               ctx.ellipse(rect.x + rect.w / 2, rect.y + rect.h / 2, rect.w / 2, rect.h / 2, 0, 0, 2 * Math.PI);
               break;
           }
           case 3: { // Heart
               let x = rect.x + rect.w / 2;
               let y = rect.y;
               let w = rect.w;
               let h = rect.h;
               let top_curve_h = h * 0.3;
               ctx.beginPath();
               ctx.moveTo(x, y + top_curve_h);
               ctx.bezierCurveTo(x, y, x - w, y, x, y + h);

               ctx.moveTo(x, y + top_curve_h);
               ctx.bezierCurveTo(x, y, x + w, y, x, y + h);
               ctx.closePath();
               break;
           }
       }
       ctx.fill();
   }

   pointToLineDist(x0, y0, x1, y1, m) {
       // Coefficients for the line in the form Ax + By + C = 0
       let A = -m;
       let B = 1;
       let C = m * x1 - y1;

       let numerator = A * x0 + B * y0 + C;
       let denominator = Math.sqrt(Math.pow(A, 2) + Math.pow(B, 2));
       return numerator / denominator;
   }
   
   envelopingLineWidth(width, height, slope) {
       let w = width / 2;
       let h = height / 2;
       let m = slope;
   
       let L = Math.sqrt((h**2 * m**2 + w**2) / (m**2 + 1));
       return L * 2; // multiply by 2 because L is the semi-length
   }

   radialDist(a, b, angle) {
       return (a * b) / Math.sqrt((a * Math.cos(angle))**2 + (b * Math.sin(angle))**2);
   }

   standardDev(V) {
       const n = V.length;
       const mean = V.reduce((a, b) => a + b) / n;
       return Math.sqrt(V.map(x => Math.pow(x - mean, 2)).reduce((a, b) => a + b) / n);
   }

   /**
    * @param {Array<number>} X length must be greater than 0 and equal to Y's
    * @param {Array<number>} Y length must be greater than 0 and equal to X's
    * @returns {{ point: { x: number, y: number }, slope: number }} A point and a slope describing the best fit line
    */
   bestFitLine(X, Y) {
       let greater_y_dev = this.standardDev(X) < this.standardDev(Y);
       if (greater_y_dev) {
           [X, Y] = [Y, X];
       }

       let sum_x = 0;
       let sum_y = 0;
       let sum_xy = 0;
       let sum_xx = 0;

       let x = 0;
       let y = 0;
       let n = X.length;

       for (let i = 0; i < n; i++) {
           x = X[i];
           y = Y[i];
           sum_x += x;
           sum_y += y;
           sum_xx += x*x;
           sum_xy += x*y;
       }

       let m, b;
       let denominator = n * sum_xx - sum_x * sum_x;
       if (denominator !== 0) {
           m = (n * sum_xy - sum_x * sum_y) / (denominator);
           b = (sum_y / n) - (m * sum_x) / n;
       } else {
           return { point: { x: X[0], y: Y[0] }, slope: 0 };
       }

       let point, slope;
       if (greater_y_dev) {
           point = { x: X[0] * m + b, y: X[0] };
           slope = 1 / m;
       } else {
           point = { x: X[0], y: X[0] * m + b };
           slope = m;
       }
       return { point, slope };
   }

   extendLine(point, slope, left, right, top, bottom) {
       let y_at_left = slope * (left - point.x) + point.y;
       let y_at_right = slope * (right - point.x) + point.y;
       let x_at_top = (top - point.y) / slope + point.x;
       let x_at_bottom = (bottom - point.y) / slope + point.x;

       var py1 = {x: x_at_top, y: top};
       var py2 = {x: x_at_bottom, y: bottom};
       var px1 = {x: left, y: y_at_left};
       var px2 = {x: right, y: y_at_right};

       var r, s;
       if (y_at_left < top) {
           r = py1;
       } else if (y_at_left <= bottom) {
           r = px1;
       } else {
           r = py2;
       }
       if (y_at_right < top) {
           s = py1;
       } else if (y_at_right <= bottom) {
           s = px2;
       } else {
           s = py2;
       }
       return [r, s];
   }

   degreesToRadians(degrees) {
       return degrees * Math.PI / 180;
   }

   parallelLinesDist(angle, x1, y1, x2, y2) {
       // Calculate the direction of the lines
       const direction_x = Math.cos(angle);
       const direction_y = Math.sin(angle);

       // Calculate the distance between the points
       const point_dist_x = x1 - x2;
       const point_dist_y = y1 - y2;

       // Return the perpendicular distance between the lines
       return Math.abs(
           (point_dist_x * direction_x + point_dist_y * direction_y) /
             Math.sqrt(direction_x ** 2 + direction_y ** 2)
       );
   }

   pointAtAngle(center_x, center_y, h_radius, v_radius, angle) {
       const x = center_x + h_radius * Math.cos(angle);
       const y = center_y + v_radius * Math.sin(angle);
       return { x, y };
   }

   prepareTextMask(ctx, cache) {
       const config = this.word_wall_configuration;
       if(config.word_pool) {
           ctx.save();

           let raw_angle = config.angle;
           if(cache['word_wall_mirror_angle']){
               raw_angle = cache['word_wall_mirror_angle'];
           }else {
               if (Math.random() < 0.5) {
                   raw_angle *= -1;
               }
               cache['word_wall_mirror_angle'] = raw_angle;
           }
           let angle = raw_angle < 0 ? raw_angle + 360 : raw_angle;
           let quarter_rotations = Math.floor(angle / 90);
           let quarter_angle_rem = angle % 90;
           let quarter_folded_angle = angle;
           quarter_folded_angle -= Math.max(0, quarter_folded_angle - 180) * 2;
           quarter_folded_angle -= Math.max(0, quarter_folded_angle - 90) * 2;

           let length = this.parallelLinesDist(this.degreesToRadians(quarter_folded_angle), 0, 0, this.natural_width, this.natural_height);
           
           let natural_size = Math.sqrt(this.natural_width * this.natural_height);
           let fontSize = Math.ceil(natural_size / 100 * config.size);
           ctx.font = "" + fontSize + "px " + config.font;

           let gap = config.horizontal_spacing;
           gap =  new Array(gap + 1).join(" ");
           let txtHeight = Math.ceil(natural_size / 100 * config.vertical_spacing);
           // idea: animated offset increase
           let offset = Math.ceil(natural_size/ 100 * config.offset);
           let textpool = [...config.word_pool];
           let txt = textpool.sort(() => 0.5 - Math.random())[0];
           if(config.draw_mode === 1){
               let r = [];
               for (let i = 0; i <  config.word_pool.length; i++) {
                   r.push(textpool.sort(() => 0.5 - Math.random())[0]);
               }
               txt = r.join(gap);
           }else if(config.draw_mode === 2){
               txt = textpool.join(gap);
           }

           let width = Math.ceil(ctx.measureText(txt + gap).width);
           offset = offset % width;

           const half_natural_width = this.natural_width / 2;
           const half_natural_height = this.natural_height/ 2;

           let center = this.pointAtAngle(half_natural_width, half_natural_height, half_natural_width, half_natural_height, this.degreesToRadians((quarter_rotations - 1) * 90));
           let radius = quarter_rotations % 2 === 0 ? half_natural_width : half_natural_height;
           let final = this.pointAtAngle(center.x, center.y, radius, radius, this.degreesToRadians(quarter_angle_rem * 2 + (quarter_rotations - 2) * 90));

           ctx.translate(final.x, final.y);
           ctx.rotate(angle * Math.PI / 180);
           ctx.fillStyle = config.color_mode === 0 ? "#000" : config.color;

           for (let i = 0; i < length; i++) {
               let w_required = length + i * offset;
               let t = txt;
               t = new Array(Math.ceil(w_required / width) + 2).join(txt + gap);
               ctx.fillText(t, -(i * offset), i * txtHeight);
           }

           ctx.restore();

           if (config.color_mode === 0) {
               ctx.globalCompositeOperation = "source-atop";
               ctx.drawImage(this.out, 0, 0, this.natural_width, this.natural_height);
           }
       }
   }

   pixelateCanvas(preset_config, cache) {
       let rect = new Rectangle(0, 0, this.natural_width, this.natural_height);

       let strength = preset_config.strength;
       if (preset_config.resolution_aware) {
           let natural_size = Math.sqrt(this.natural_width * this.natural_height);
           strength = Math.max(1, Math.round(natural_size / 1000 * preset_config.strength));
       }

       let color_precision = 256 / (preset_config.channel_colors - 1);
       if (preset_config.type === 1) {
           if (0 < preset_config.spacing) {
               this.auxctx.reset();
               hexagonalPixel(this.auxctx, rect, strength, preset_config.spacing, this.clip, this.clipctx, preset_config.grayscale, color_precision, preset_config.accurate_sampling);
               this.outctx.drawImage(this.aux, 0, 0);
           } else {
               hexagonalPixel(this.outctx, rect, strength, preset_config.spacing, this.clip, this.clipctx, preset_config.grayscale, color_precision, preset_config.accurate_sampling);
           }
       } else if (preset_config.type === 2) {
           glitchedPixel(this.outctx, rect, strength, preset_config.spacing, this.clip, this.clipctx, preset_config.grayscale, color_precision, preset_config.accurate_sampling);
       } else {
           mosaicPixel(this.outctx, rect, strength, preset_config.spacing, this.clip, this.clipctx, preset_config.grayscale, color_precision, preset_config.accurate_sampling);
       }

       this.finishReverse();
   }

   pixelateAreas(results, cache) {
       const putPixels = (rect, strength, preset_config) => {
           let color_precision = 256 / (preset_config.channel_colors - 1);
           if (preset_config.type === 1) {
               if (0 < preset_config.spacing) {
                   hexagonalPixel(this.auxctx, rect, strength, preset_config.spacing, this.clip, this.clipctx, preset_config.grayscale, color_precision, preset_config.accurate_sampling, "source-over");
                   this.fillctx.globalCompositeOperation = "source-atop";
                   this.fillctx.drawImage(this.out, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
                   this.fillctx.globalCompositeOperation = "source-atop";
                   this.fillctx.drawImage(this.aux, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
               } else {
                   hexagonalPixel(this.fillctx, rect, strength, preset_config.spacing, this.clip, this.clipctx, preset_config.grayscale, color_precision, preset_config.accurate_sampling, "source-atop");
               }
           } else if (preset_config.type === 2) {
               this.fillctx.globalCompositeOperation = "source-atop";
               this.fillctx.drawImage(this.out, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
               glitchedPixel(this.fillctx, rect, strength, preset_config.spacing, this.clip, this.clipctx, preset_config.grayscale, color_precision, preset_config.accurate_sampling, "source-atop");
           } else {
               if (0 < preset_config.spacing) {
                   this.fillctx.globalCompositeOperation = "source-atop";
                   this.fillctx.drawImage(this.out, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
               }
               mosaicPixel(this.fillctx, rect, strength, preset_config.spacing, this.clip, this.clipctx, preset_config.grayscale, color_precision, preset_config.accurate_sampling, "source-atop");
           }
       }

       let results_by_preset_ident = Object.groupBy(results, (res) => res.preset_ident);
       for (let preset_ident in results_by_preset_ident) {
           let results = results_by_preset_ident[preset_ident];
           let preset_config = results[0].preset_config;

           let base_strength = preset_config.strength;
           let natural_size = Math.sqrt(this.natural_width * this.natural_height);
           if (preset_config.resolution_aware) {
               base_strength = natural_size / 1000 * preset_config.strength;
           }

           if (preset_config.type === 1 && 0 < preset_config.spacing) {
               this.auxctx.reset();
           }

           if (preset_config.shape === 4) {
               this.handleLineShapes(results, preset_config, cache, (rect) => {
                   let strength = Math.max(1, Math.round(base_strength));
                   putPixels(rect, strength, preset_config);

                   this.finishFill(rect, cache);
               })
           } else {
               this.handleShapes(results, preset_config, cache, (prefeather_rect, rect) => {
                   let strength;
                   if (preset_config.detection_size_aware) {
                       let threshold = natural_size / 4;
                       let rect_size = Math.sqrt(prefeather_rect.w * prefeather_rect.h);
                       let ratio = rect_size / threshold;
                       strength = Math.max(1, Math.round(base_strength * ratio));
                   } else {
                       strength = Math.max(1, Math.round(base_strength));
                   }

                   putPixels(rect, strength, preset_config);
               }, null)
           }
       }
   }

   blurCanvas(preset_config, cache) {
       let strength = preset_config.strength;
       if (preset_config.resolution_aware) {
           let natural_size = Math.sqrt(this.natural_width * this.natural_height);
           strength = Math.max(1, Math.round(natural_size / 1000 * preset_config.strength));
       }

       let blurred = generateBaseBlurImproved(this.clip, strength, preset_config.grayscale);
       this.outctx.drawImage(blurred, 0, 0, this.natural_width, this.natural_height, 0, 0, this.natural_width, this.natural_height);

       this.finishReverse();
   }

   blurAreas(results, cache) {
       let results_by_preset_ident = Object.groupBy(results, (res) => res.preset_ident);
       for (let preset_ident in results_by_preset_ident) {
           let results = results_by_preset_ident[preset_ident];
           let preset_config = results[0].preset_config;
           
           let strength = preset_config.strength;
           if (preset_config.resolution_aware) {
               let natural_size = Math.sqrt(this.natural_width * this.natural_height);
               strength = Math.round(natural_size / 1000 * preset_config.strength);
           }
           strength = Math.max(1, strength);

           this.auxctx.reset();
           let blurred = generateBaseBlurImproved(this.clip, strength, preset_config.grayscale);

           if (preset_config.shape === 4) {
               this.handleLineShapes(results, preset_config, cache, (rect) => {
                   this.fillctx.globalCompositeOperation = "source-atop";
                   this.fillctx.drawImage(blurred, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);

                   this.finishFill(rect, cache);
               });
           } else {
               this.handleShapes(results, preset_config, cache, (prefeather_rect, rect) => {
                   this.fillctx.globalCompositeOperation = "source-atop";
                   this.fillctx.drawImage(blurred, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
               }, null);
           }
       }
   }

   triangulateCanvas(preset_config, cache) {
       let vertex_count = preset_config.vertex_count;
       if (preset_config.resolution_aware) {
           vertex_count = Math.round(vertex_count * Math.sqrt(this.natural_width * this.natural_height) / 1024);
       }
       let natural_size = this.natural_width * this.natural_height;
       if (natural_size <= 1) return;
       vertex_count = clamp(vertex_count, 1, natural_size);

       const fill = preset_config.fill_color_mode === 0 ? true
           // Assume fill_color_mode === 1
           : isHexColorTransparent(preset_config.fill_color) ? false
           : preset_config.fill_color;
       const stroke = preset_config.stroke_color_mode === 0 ? true
           // Assume stroke_color_mode === 1
           : isHexColorTransparent(preset_config.stroke_color) ? false
           : preset_config.stroke_color;
       let triangulation_params = {
           accuracy: preset_config.accuracy,
           // TODO: make resolution aware
           blur: Math.round(preset_config.blurring),
           threshold: preset_config.threshold,
           vertexCount: vertex_count,
           fill: fill,
           stroke: stroke,
           strokeWidth: Math.max(0.1, Math.sqrt(this.natural_width * this.natural_height) / 2048) * preset_config.stroke_width,
           gradients: preset_config.gradients,
           gradientStops: preset_config.gradient_stops,
           lineJoin: ['miter', 'round', 'bevel'][preset_config.line_join],
           transparentColor: preset_config.transparent_color,
       };

       let image_data = this.clipctx.getImageData(0, 0, this.natural_width, this.natural_height);
       const triangulized_image_data = triangulate(triangulation_params)
           .fromImageDataSync(image_data)
           .toImageDataSync();
       this.outctx.putImageData(triangulized_image_data, 0, 0);
       this.finishReverse();
   }

   triangulateAreas(results, cache) {
       const putTriangles = (rect, preset_config) => {
           let rect_coverage = (rect.w * rect.h) / (this.natural_width * this.natural_height);

           let vertex_count;
           if (!preset_config.resolution_aware) {
               vertex_count = Math.round(preset_config.vertex_count * rect_coverage);
           } else {
               vertex_count = Math.round(preset_config.vertex_count * rect_coverage * Math.sqrt(this.natural_width * this.natural_height) / 1.024);
           }
           let size = rect.w * rect.h;
           if (size <= 1) return;
           vertex_count = clamp(vertex_count, 1, size);

           const fill = preset_config.fill_color_mode === 0 ? true
               // Assume fill_color_mode === 1
               : isHexColorTransparent(preset_config.fill_color) ? false
               : preset_config.fill_color;
           const stroke = preset_config.stroke_color_mode === 0 ? true
               // Assume stroke_color_mode === 1
               : isHexColorTransparent(preset_config.stroke_color) ? false
               : preset_config.stroke_color;
           let triangulation_params = {
               accuracy: preset_config.accuracy,
               // TODO: make resolution aware
               blur: Math.round(preset_config.blurring),
               threshold: preset_config.threshold,
               vertexCount: vertex_count,
               fill: fill,
               stroke: stroke,
               strokeWidth: Math.max(0.1, Math.sqrt(this.natural_width * this.natural_height) / 2048) * preset_config.stroke_width,
               gradients: preset_config.gradients,
               gradientStops: preset_config.gradient_stops,
               lineJoin: ['miter', 'round', 'bevel'][preset_config.line_join],
               transparentColor: preset_config.transparent_color,
           };

           let image_data = this.clipctx.getImageData(rect.x, rect.y, rect.w, rect.h);
           const triangulized_image_data = triangulate(triangulation_params)
               .fromImageDataSync(image_data)
               .toImageDataSync();
           this.auxctx.putImageData(triangulized_image_data, rect.x, rect.y);
       }

       let results_by_preset_ident = Object.groupBy(results, (res) => res.preset_ident);
       for (let preset_ident in results_by_preset_ident) {
           let results = results_by_preset_ident[preset_ident];
           let preset_config = results[0].preset_config;
           
           if (preset_config.shape === 4) {
               this.handleLineShapes(results, preset_config, cache, (rect) => {
                   this.auxctx.reset();
                   putTriangles(rect.scaled(1.05, 1.05), preset_config);
                   this.fillctx.globalCompositeOperation = "source-atop";
                   this.fillctx.drawImage(this.aux, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);

                   this.finishFill(rect, cache);
               });
           } else {
               this.handleShapes(results, preset_config, cache, null, (rect) => {
                   this.auxctx.reset();
                   putTriangles(rect.scaled(1.05, 1.05), preset_config);
                   this.fillctx.globalCompositeOperation = "source-atop";
                   this.fillctx.drawImage(this.aux, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);

                   this.finishFill(rect, cache);
               });
           }
       }
   }

   barOnCanvas(preset_config, cache) {
       if (preset_config.type === 0) {
           this.outctx.fillStyle = preset_config.color;
           this.outctx.globalCompositeOperation = "source-over";
       } else {
           this.outctx.globalAlpha = preset_config.opacity;
           this.outctx.globalCompositeOperation = "destination-out";
       }
       this.outctx.fillRect(0, 0, this.natural_width, this.natural_height);

       this.finishReverse();
   }

   barOnAreas(results, cache) {
       const paintBars = (rect, preset_config) => {
           if (preset_config.type === 0) {
               this.fillctx.globalCompositeOperation = "source-atop";
               this.fillctx.fillStyle = preset_config.color;
               this.fillctx.fillRect(rect.x, rect.y, rect.w, rect.h);

               this.finishFill(rect, cache);
           } else { 
               this.outctx.globalCompositeOperation = "destination-out";
               this.outctx.drawImage(this.fill, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);

               // Custom finish fill
               this.outctx.globalAlpha = 1 - preset_config.opacity;
               this.outctx.globalCompositeOperation = "source-over";
               this.outctx.drawImage(this.fill, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);

               this.outctx.globalAlpha = 1;
               if (!cache.reverse_mode && this.word_wall_configuration.enabled) {
                   this.auxctx.restore();

                   this.auxctx.globalCompositeOperation = "source-over";
                   this.auxctx.drawImage(this.fill, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
                   this.auxctx.globalCompositeOperation = "source-in";
                   this.auxctx.drawImage(this.ww, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
                   this.outctx.globalCompositeOperation = "source-over";
                   this.outctx.drawImage(this.aux, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
               }
           }
       }

       let results_by_preset_ident = Object.groupBy(results, (res) => res.preset_ident);
       for (let preset_ident in results_by_preset_ident) {
           let results = results_by_preset_ident[preset_ident];
           let preset_config = results[0].preset_config;

           if (preset_config.shape === 4) {
               this.handleLineShapes(results, preset_config, cache, (rect) => {
                   this.fillctx.globalCompositeOperation = "source-atop";
                   this.fillctx.drawImage(this.clip, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);

                   paintBars(rect, preset_config);
               });
           } else {
               this.handleShapes(results, preset_config, cache, (prefeather_rect, rect) => {
                   this.fillctx.globalCompositeOperation = "source-atop";
                   this.fillctx.drawImage(this.clip, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
               }, (rect) => {
                   paintBars(rect, preset_config);
               });
           }
       }
   }

   sobelCanvasPrewitt(preset_config, cache) {
       let sobel_image_data = this.gradient(this.clipctx.getImageData(0, 0, this.natural_width, this.natural_height));
       if (preset_config.inverted_color) {
           const temp_canvas = document.createElement("canvas");
           temp_canvas.width = this.natural_width;
           temp_canvas.height = this.natural_height;
           const temp_ctx = temp_canvas.getContext("2d");
           temp_ctx.putImageData(sobel_image_data, 0, 0);
           temp_ctx.globalCompositeOperation = "exclusion";
           temp_ctx.fillStyle = "white";
           temp_ctx.fillRect(0, 0, this.natural_width, this.natural_height);
           this.outctx.drawImage(temp_canvas, 0, 0, this.natural_width, this.natural_height);
       } else {
           this.outctx.putImageData(sobel_image_data, 0, 0);
       }

       this.finishReverse();
   }

   sobelAreasPrewitt(results, cache) {
       let results_by_preset_ident = Object.groupBy(results, (res) => res.preset_ident);
       for (let preset_ident in results_by_preset_ident) {
           let results = results_by_preset_ident[preset_ident];
           let preset_config = results[0].preset_config;
           
           this.auxctx.reset();
           let sobel_image_data = this.gradient(this.clipctx.getImageData(0, 0, this.natural_width, this.natural_height));
           if (preset_config.inverted_color) {
               const temp_canvas = document.createElement("canvas");
               temp_canvas.width = this.natural_width;
               temp_canvas.height = this.natural_height;
               const temp_ctx = temp_canvas.getContext("2d");
               temp_ctx.putImageData(sobel_image_data, 0, 0);
               temp_ctx.globalCompositeOperation = "exclusion";
               temp_ctx.fillStyle = "white";
               temp_ctx.fillRect(0, 0, this.natural_width, this.natural_height);
               this.auxctx.drawImage(temp_canvas, 0, 0, this.natural_width, this.natural_height);
           } else {
               this.auxctx.putImageData(sobel_image_data, 0, 0);
           }

           if (preset_config.shape === 4) {
               this.handleLineShapes(results, preset_config, cache, (rect) => {
                   this.fillctx.globalCompositeOperation = 'source-atop';
                   this.fillctx.drawImage(this.aux, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);

                   this.finishFill(rect, cache);
               });
           } else {
               this.handleShapes(results, preset_config, cache, (prefeather_rect, rect) => {
                   this.fillctx.globalCompositeOperation = 'source-atop';
                   this.fillctx.drawImage(this.aux, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
               }, null);
           }
       }
   }

   splatterCanvas(preset_config, cache) {
       if (cache.mixed_mode) {
           this.splatter(this.outctx, true, new Rectangle(0, 0, this.natural_width, this.natural_height), preset_config);
       } else {
           this.auxctx.reset();
           this.splatter(this.auxctx, true, new Rectangle(0, 0, this.natural_width, this.natural_height), preset_config);
           this.outctx.drawImage(this.aux, 0, 0);
       }

       this.finishReverse();
   }

   clipAreasWithSplatter(results, cache) {
       this.fillctx.reset();

       for (const res of results) {
           let { rect, preset_config } = res;
           rect = rect.scaled(preset_config.scale, preset_config.scale);
           this.splatter(this.fillctx, false, rect, preset_config);
       }

       this.fillctx.globalCompositeOperation = "source-atop";
       this.fillctx.drawImage(this.clip, 0, 0);
       this.outctx.drawImage(this.fill, 0, 0);
   }

   splatterAreas(results, cache) {
       this.fillctx.reset();

       for (const res of results) {
           let { rect, preset_config } = res;
           rect = rect.scaled(preset_config.scale, preset_config.scale);
           this.splatter(this.fillctx, true, rect, preset_config);
       }

       this.finishFill(new Rectangle(0, 0, this.natural_width, this.natural_height), cache);
   }

   sampleColor(clipctx, x, y, preset_config) {
       const pixelArr = clipctx.getImageData(x, y, 1, 1).data;
       return "rgba(" + pixelArr[0] + "," + pixelArr[1] + "," + pixelArr[2] + "," + preset_config.opacity + ")";
   }

   subSplatter(ctx, colored, parent_x, parent_y, parent_size, size_modifier, angle, remaining_iterations, preset_config) {
       remaining_iterations--;
       
       let color;
       if (colored) {
           if (preset_config.color_mode === 1) {
               // TODO: cache weights?
               const weights = preset_config.color_pool.map(item => item.weight);
               color = preset_config.color_pool[randomWeighted(weights)].color;
           }
       } else {
           color = "#000";
       }

       for (let i = 0; i < preset_config.sub_amount; i += 1) {
           let dist = Math.floor(Math.random() * parent_size * 0.8 + parent_size);
           angle = this.angleRND(angle);
           let rect_size = (parent_size * 1.5 - dist * 0.5) * preset_config.sub_size;
           let x = (parent_x + dist * Math.cos(angle)) - rect_size / 2;
           let y = (parent_y + dist * Math.sin(angle)) - rect_size / 2;
           
           if (rect_size <= 2) {
               continue;
           }

           let d = (rect_size * 2) / size_modifier;
           let size = Math.floor(Math.random() * (d - 1)) + 1;
           
           ctx.beginPath();
           ctx.fillStyle = color ?? this.sampleColor(this.clipctx, x, y, preset_config);
           ctx.arc(x, y, size, 0, 2 * Math.PI);
           ctx.fill();

           if (0 < remaining_iterations) {
               this.subSplatter(ctx, colored, x, y, size, size_modifier, angle, remaining_iterations, preset_config)
           }
       }
   }

   splatter(ctx, colored, rect, preset_config) {
       let size_modifier = 8 / preset_config.size;
       let splatters_num = Math.ceil(Math.sqrt(rect.w * rect.h) / 10 * preset_config.amount);
       let iterations = preset_config.recursions;

       let color;
       if (colored) {
           if (preset_config.color_mode === 1) {
               // TODO: cache weights?
               const weights = preset_config.color_pool.map(item => item.weight);
               color = preset_config.color_pool[randomWeighted(weights)].color;
           }
       } else {
           color = "#000";
       }
       
       for (let i = 0; i < splatters_num; i++) {
           let [x, y] = this.centeredRND(rect.x, rect.w, rect.y, rect.h, preset_config.centering);
           let d = (rect.w + rect.h) / size_modifier;
           let size = Math.floor(Math.random() * (d - 1)) + 1;

           ctx.beginPath();
           ctx.fillStyle = color ?? this.sampleColor(this.clipctx, x, y, preset_config);
           ctx.arc(x, y, size, 0, 2 * Math.PI);
           ctx.fill();

           if (0 < iterations) {
               this.subSplatter(ctx, colored, x, y, size, size_modifier, null, iterations, preset_config);
           }
       }
   }

   centeredRND(x, w, y, h, wdh) {
       let rndY = Math.floor(Math.random() * (h)) + y;
       let rndX = Math.floor(Math.random() * (w)) + x;
       if (wdh > 0) {
           for (let i = 0; i < wdh - 1; i++) {
               rndY = (rndY + Math.floor(Math.random() * (h)) + y);
               rndX = (rndX + Math.floor(Math.random() * (w)) + x);
           }
           rndY /= wdh;
           rndX /= wdh;
           rndY = Math.floor(rndY);
           rndX = Math.floor(rndX);
       }
       return [rndX, rndY];
   }

   angleRND(angle) {
       let new_angle;
       if (angle) {
           new_angle = angle + ((Math.round(Math.random()) * 2 - 1) * Math.floor(Math.random() * (90)));
       } else {
           new_angle = Math.floor(Math.random() * (360));
       }
       return new_angle;
   }

   conv3x(data, idx, w, m) {
       return (m[0] * data[idx - w - 4] + m[1] * data[idx - 4] + m[2] * data[idx + w - 4] - m[0] * data[idx - w + 4] - m[1] * data[idx + 4] - m[2] * data[idx + 4 + 4]);
   }

   conv3y(data, idx, w, m) {
       return (m[0] * data[idx - w - 4] + m[1] * data[idx - w] + m[2] * data[idx - w + 4] - (m[0] * data[idx + w - 4] + m[1] * data[idx + w] + m[2] * data[idx + w + 4]));
   }

   gradient_internal(pixels, mask) {
       var data = pixels.data;
       var w = pixels.width * 4;
       var l = data.length - w - 4;
       var buff = new data.constructor(new ArrayBuffer(data.length));

       for (var i = w + 4; i < l; i += 4) {
           var dx = this.conv3x(data, i, w, mask);
           var dy = this.conv3y(data, i, w, mask);
           buff[i] = buff[i + 1] = buff[i + 2] = Math.sqrt(dx * dx + dy * dy);
           buff[i + 3] = 255;
       }
       pixels.data.set(buff);
   }

   gradient(imagedata) {
       this.gradient_internal(imagedata, [1, 2, 1]); // Apply Sobel operator
       return imagedata;
   }

   /**
    * Inverts the colors of a given image(data).
    * ! Good Performance in Background, BAD PERFORMANCE ON CONTENT SCRIPT / VIDEOS
    * @param imgData - imageData Object
    */
   invertImageData(imgData) {
       for (let i = 0; i < imgData.data.length; i += 4) {
           let r = imgData.data[i];
           let g = imgData.data[i + 1];
           let b = imgData.data[i + 2];
           r = 255 - r;
           g = 255 - g;
           b = 255 - b;
           imgData.data[i] = r;
           imgData.data[i + 1] = g;
           imgData.data[i + 2] = b;
           imgData.data[i + 3] = 255;
       }
   }

   stickerOnCanvas(preset_config, preset_ident, cache) {
       let rect = new Rectangle(0, 0, this.natural_width, this.natural_height);
       let sticker = this.getSticker(preset_config, preset_ident, {}, rect, cache);
       if (sticker == null) return;

       rect = this.paintSticker(this.outctx, rect, sticker);

       this.finishReverse();
   }

   clipAreasWithSticker(results, cache) {
       this.fillctx.reset();
       
       for (const res of results) {
           let { klass, rect, preset_config, preset_ident } = res;
           rect = rect.scaled(preset_config.scale, preset_config.scale);

           let sticker = this.getSticker(preset_config, preset_ident, klass, rect, cache);
           if (sticker == null) continue;

           this.fillctx.globalCompositeOperation = "source-over";
           rect = this.paintSticker(this.fillctx, rect, sticker);
           this.fillctx.globalCompositeOperation = "source-atop";
           this.fillctx.drawImage(this.clip, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);

           this.outctx.drawImage(this.fill, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
       }
   }

   stickerOnAreas(results, cache) {
       this.fillctx.reset();
       
       for (const res of results) {
           let { klass, rect, preset_config, preset_ident } = res;
           rect = rect.scaled(preset_config.scale, preset_config.scale);

           let sticker = this.getSticker(preset_config, preset_ident, klass, rect, cache);
           if (sticker == null) continue;

           this.paintSticker(this.fillctx, rect, sticker);
       }

       this.finishFill(new Rectangle(0, 0, this.natural_width, this.natural_height), cache);
   }

   paintSticker(ctx, rect, sticker) {
       rect = rect.scaled(sticker.scale, sticker.scale);
       let buffer = sticker.file;
       const sticker_canvas = document.createElement("canvas");
       sticker_canvas.width = sticker.width;
       sticker_canvas.height = sticker.height;
       const sticker_ctx = sticker_canvas.getContext("2d");
       let sticker_data = sticker_ctx.createImageData(sticker_canvas.width, sticker_canvas.height);
       if (buffer != null) {
           sticker_data.data.set(buffer);
       } else {
           console.error("NULL Data for Sticker", sticker);
       }

       sticker_ctx.putImageData(sticker_data, 0, 0);
       let h_ratio = rect.w / sticker_canvas.width;
       let v_ratio = rect.h / sticker_canvas.height;
       let ratio = Math.max(h_ratio, v_ratio);
       let w_adjusted = sticker_canvas.width * ratio;
       let h_adjusted = sticker_canvas.height * ratio;
       let x_adjusted = rect.x - (Math.abs(rect.w - w_adjusted) / 2);
       let y_adjusted = rect.y - (Math.abs(rect.h - h_adjusted) / 2);
       rect = new Rectangle(x_adjusted, y_adjusted, w_adjusted, h_adjusted);

       ctx.drawImage(sticker_canvas, rect.x, rect.y, rect.w, rect.h);
       return rect;
   }

   getSticker(preset_config, preset_ident, klass, rect, cache) {
       if (preset_config.draw_mode === 0 && cache[`preset_ident_${preset_ident}_sticker_for_${klass.key}`] != null) {
           return cache[`preset_ident_${preset_ident}_sticker_for_${klass.key}`];
       }

       let stickers = [];
       if (cache[`preset_ident_${preset_ident}_stickers`] != null) {
           stickers = cache[`preset_ident_${preset_ident}_stickers`];
       } else {
           let source_names = preset_config.enabled_sources;
           if (preset_config.sample_single_source) {
               source_names = [source_names[Math.floor(Math.random() * source_names.length)]];

               let collection = this.sticker_collections.find(collection => source_names[0] === collection._name);
               if (collection != null) {
                   stickers = collection._stickers;
               } else {
                   this.sticker_collections.forEach((collection) => {
                       stickers.push(...getStickersByGroup(collection, source_names));
                   });
               }
           } else {
               this.sticker_collections.forEach(collection => {
                   if (source_names.some(name => collection._name === name)) {
                       stickers.push(...collection._stickers);
                   } else {
                       stickers.push(...getStickersByGroup(collection, source_names));
                   }
               });
           }

           cache[`preset_ident_${preset_ident}_stickers`] = stickers;
       }

       // NOTE: Basically if not reverse
       // TODO: there is a chance of drawing a single group or collection that doesn't contain any stickers for one given klass, what should we do then? currently, just nothing is drawn
       if (klass.index >= 0) {
           stickers = stickers.filter(sticker => {
               return sticker.klasses.some(other_klass => other_klass.key === klass.key);
           });
       }

       if (stickers.length === 0) {
           return null;
       }
       
       let sticker = this.randomSticker(stickers);
       if (preset_config.draw_mode == 0) {
           cache[`preset_ident_${preset_ident}_sticker_for_${klass.key}`] = sticker;
       }

       return sticker;
   }

   randomSticker(stickers) {
       if (stickers.length === 0) {
           return null;
       }

       let i;
       let weights = [];
       for (i = 0; i < stickers.length; i++) {
           if (!stickers[i].chance && stickers[i].chance !== 0) {
               stickers[i].chance = 1;
           }
           weights[i] = stickers[i].chance + (weights[i - 1] || 0);
       }
       let random = Math.random() * weights[weights.length - 1];
       for (i = 0; i < weights.length; i++) {
           if (weights[i] > random) {
               break;
           }
       }
       return stickers[i];
   }

   computeFontSizeToFitWidth(ctx, text, max_width, family, min_size, max_size) {
       ctx.font = `${max_size}px ${family}`;
       if (Math.ceil(ctx.measureText(text).width) <= max_width) {
           return max_size;
       }

       let low = min_size;
       let high = max_size;
       let lowest_fit = min_size;
       while (low <= high) {
           const size = Math.floor((low + high) / 2);
           ctx.font = `${size}px ${family}`;
           const width = Math.ceil(ctx.measureText(text).width);
           
           if (max_width < width) {
               high = size - 1;
           } else {
               low = size + 1;
               lowest_fit = size;
           }
       }
     
       return lowest_fit;
   }

   boxOnCanvas(preset_config, cache) {
       this.outctx.fillStyle = preset_config.fill_color;
       this.outctx.fillRect(0, 0, this.natural_width, this.natural_height);
       this.finishReverse();
   }

   clipAreasWithBox(results, cache) {
       this.fillctx.reset();
       // For a border width of 2px on a 1024x1024px pic, ensure it being a multiple of 2
       let border_width = Math.max(2, Math.round(2 * Math.sqrt(this.natural_width * this.natural_height) / 2048) * 2);
       results.sort((a, b) => {
           if (a.rect.y < b.rect.y) return -1; 
           else if (b.rect.y < a.rect.y) return 1;
           else return 0
       })
       let [no_fill_results, fill_results] = results.reduce((output, item) => {
           if (cache.labels.includes(item.klass.key) && item.preset_config.fill === 0) {
               output[1].push(item);
           } else {
               output[0].push(item);
           }
           return output;
       }, [[], []]);
       for (const res of no_fill_results) {
           let { rect, preset_config, klass, score } = res;
           rect = rect.scaled(preset_config.scale, preset_config.scale);

           this.fillctx.drawImage(this.clip, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);

           this.drawBoxBorder(this.fillctx, border_width, rect, klass, score, preset_config);

           this.outctx.drawImage(this.fill, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
       }
       for (const res of fill_results) {
           let { rect, preset_config, klass, score } = res;
           rect = rect.scaled(preset_config.scale, preset_config.scale);

           const fill_color = preset_config.fill_color;
           this.fillctx.fillStyle = fill_color;
           if (!isHexColorOpaque(fill_color)) {
               this.fillctx.drawImage(this.clip, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
           }
           this.fillctx.fillRect(rect.x, rect.y, rect.w, rect.h);

           this.drawBoxBorder(this.fillctx, border_width, rect, klass, score, preset_config);

           this.outctx.drawImage(this.fill, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
       }
   }

   boxOnAreas(results, cache) {
       this.auxctx.reset();
       this.fillctx.reset();

       // For a border width of 2px on a 1024x1024px pic, ensure it being a multiple of 2
       let border_width = Math.max(2, Math.round(2 * Math.sqrt(this.natural_width * this.natural_height) / 2048) * 2);

       results.sort((a, b) => {
           if (a.rect.y < b.rect.y) return -1; 
           else if (b.rect.y < a.rect.y) return 1;
           else return 0
       })
       let [no_fill_results, fill_results] = results.reduce((output, item) => {
           if (cache.labels.includes(item.klass.key) && item.preset_config.fill === 0) {
               output[1].push(item);
           } else {
               output[0].push(item);
           }
           return output;
       }, [[], []]);
       for (const res of no_fill_results) {
           let { rect, preset_config, klass, score } = res;
           rect = rect.scaled(preset_config.scale, preset_config.scale);

           this.fillctx.drawImage(this.clip, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);

           this.drawBoxBorder(this.auxctx, border_width, rect, klass, score, preset_config);
       }
       for (const res of fill_results) {
           let { rect, preset_config, klass, score } = res;
           rect = rect.scaled(preset_config.scale, preset_config.scale);

           const fill_color = preset_config.fill_color;
           this.fillctx.fillStyle = fill_color;
           if (!isHexColorOpaque(fill_color)) {
               this.fillctx.drawImage(this.clip, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
           }
           this.fillctx.fillRect(rect.x, rect.y, rect.w, rect.h);

           this.drawBoxBorder(this.auxctx, border_width, rect, klass, score, preset_config);
       }

       this.finishFill(new Rectangle(0, 0, this.natural_width, this.natural_height), cache);
       this.outctx.drawImage(this.aux, 0, 0, this.natural_width, this.natural_height);
   }

   drawBoxBorder(ctx, border_width, rect, klass, score, preset_config) {
       const border_color = preset_config.border_color;
       if (!isHexColorTransparent(border_color)) {
           ctx.strokeStyle = border_color;
           ctx.lineWidth = border_width;
           ctx.strokeRect(rect.x, rect.y, rect.w, rect.h);

           let min_font_size = 2 + Math.ceil((this.natural_width * 7) / 1000);
           let max_font_size = 2 + Math.ceil((this.natural_width * 14) / 1000);
           min_font_size = Math.max(8, min_font_size);
           max_font_size = Math.max(8, max_font_size);

           if (preset_config.display === 0 || preset_config.display === 1) {
               let text_to_write = klass.name;
               this.drawTextBox(ctx, border_width, rect.x, rect.y, rect.w, rect.h, text_to_write, border_color, "top", min_font_size, max_font_size);
           }
           if (preset_config.display === 0 || preset_config.display === 2) {
               let text_to_write = score;
               this.drawTextBox(ctx, border_width, rect.x, rect.y, rect.w, rect.h, text_to_write, border_color, "bottom", min_font_size - 2, max_font_size - 2);
           }
       }
   }

   drawTextBox(ctx, padding, x, y, width, height, text, bg_color, y_baseline = "top", min_font_size = 8, max_font_size = 16, font_family = "sans-serif", fg_color = "#000000") {
       ctx.textBaseline = y_baseline;
       
       const box_padding = padding / 2;
       const text_padding = padding;
       const max_string_width = width - box_padding * 2 - text_padding * 2;

       let font_size = this.computeFontSizeToFitWidth(ctx, text, max_string_width, font_family, min_font_size, max_font_size);
       ctx.font = `${font_size}px ${font_family}`;
       const string_width = Math.ceil(ctx.measureText(text).width);
       if (max_string_width < string_width) return;

       let box_x, text_x;
       box_x = x + box_padding;
       text_x = box_x + text_padding;

       let box_y, text_y;
       if (y_baseline === "top") {
           box_y = y + box_padding;
           text_y = box_y + text_padding;
       } else { // Assumed to be "bottom"
           box_y = y + height - box_padding - text_padding * 2 - font_size;
           text_y = box_y + text_padding + font_size;
       }

       ctx.fillStyle = bg_color;
       let box_width = string_width + text_padding * 2;
       let box_height = font_size + text_padding * 2;
       ctx.fillRect(box_x, box_y, box_width, box_height);

       // text = text + `sz: ${string_width}`;
       ctx.fillStyle = fg_color;
       ctx.fillText(text, text_x, text_y);
   }

   neppyfiArea(klass, rect, preset_config, cache) {
       const dummyConfig = {};
       // The amount of layers that will be generated.
       dummyConfig.number_of_layers = preset_config.multiple_panels_amount;

       // The fill of these layers
       /*
       dummyConfig.gradient_configuration = [
           {"offset": 0.05, "color": "#63A3FF5F"},
           {"offset": 0.2, "color": "#B6DCFFCF"},
           {"offset": 0.3, "color": "#B6DCFFCF"},
           {"offset": 0.5, "color": "#72BEFF5F"},
       ]*/
       //dummyConfig.gradient_configuration = [{"offset": 0.0, "color": "#0000007F"},];
       dummyConfig.gradient_configuration = preset_config.multiple_panels_gradient;
       dummyConfig.is_border_enabled = preset_config.multiple_panels_border;
       // Panes will not be smaller than min_size times the size of the detection area.
       dummyConfig.min_size = preset_config.multiple_panels_min_size;

       // Panes will be at most max_size times the size of the detection area.
       // NOTE: Similar to regular scale, but should default to like 1.5 instead of 1.
       dummyConfig.max_size = preset_config.scale+(preset_config.multiple_panels_max_size-1.0);//bar_configuration.scale + 0.45;

       // How far are the layers allowed to scatter?
       // NOTE: Range from 0.0 upwards. As a factor of the detection range.
       dummyConfig.scatter = preset_config.multiple_panels_scatter;

       // Probability of a pair of boxes appearing instead of one box.
       dummyConfig.split_chance = preset_config.multiple_panels_split_chance;

       // NOTE: Why was this??
       // if (Math.max(rect.w, rect.h) < 600) {
       //     dummyConfig.number_of_layers -= 1;
       // }
       if (klass.erotic) {
           dummyConfig.number_of_layers += 1;
       }


       let random_width_scale = 1;
       let random_height_scale = 1;
       let x_shift = 0;
       let y_shift = 0;

       for (let i = 0; i < dummyConfig.number_of_layers; i++) {

           if (i > 0 && Math.random() < dummyConfig.split_chance) {
               this.neppySplit(rect, dummyConfig);
           } else {
               // Regular layer
               const random_rect = rect.scaled(random_width_scale, random_height_scale)
                   .shifted(x_shift, y_shift)
                   .global_cropped(0, 0, this.natural_width, this.natural_height);
               this.neppyFiOnce(random_rect, dummyConfig);
           }

           random_width_scale = generate_random_float_in_range(dummyConfig.min_size, dummyConfig.max_size);
           random_height_scale = generate_random_float_in_range(dummyConfig.min_size, dummyConfig.max_size);
           x_shift = generate_random_in_range(-rect.w * dummyConfig.scatter, rect.w * dummyConfig.scatter);
           y_shift = generate_random_in_range(-rect.h * dummyConfig.scatter, rect.h * dummyConfig.scatter);
       }
   }

   neppySplit(rect, dummyConfig) {
       // Split layer
       let rect1, rect2;
       if (Math.random() < 0.5) {
           // Horizontal split
           rect1 = rect.scaled(1, 0.5);
           rect2 = rect1.shifted(0, rect1.h / 2);
           rect1 = rect1.shifted(0, -rect1.h / 2);
       } else {
           // Vertical split
           rect1 = rect.scaled(0.5, 1);
           rect2 = rect1.shifted(rect1.w / 2, 0);
           rect1 = rect1.shifted(-rect1.w / 2, 0);
       }

       let x_shift = generate_random_in_range(-rect1.w * dummyConfig.scatter, rect1.w * dummyConfig.scatter);
       let y_shift = generate_random_in_range(-rect1.h * dummyConfig.scatter, rect1.h * dummyConfig.scatter);
       let random_width_scale = generate_random_float_in_range(dummyConfig.min_size, dummyConfig.max_size);
       let random_height_scale = generate_random_float_in_range(dummyConfig.min_size, dummyConfig.max_size);

       rect1 = rect1.scaled(random_width_scale, random_height_scale)
           .shifted(x_shift, y_shift)
           .global_cropped(0, 0, this.natural_width, this.natural_height);

       this.neppyFiOnce(rect1, dummyConfig);

       x_shift = generate_random_in_range(-rect2.w * dummyConfig.scatter, rect2.w * dummyConfig.scatter);
       y_shift = generate_random_in_range(-rect2.h * dummyConfig.scatter, rect2.h * dummyConfig.scatter);
       random_width_scale = generate_random_float_in_range(dummyConfig.min_size, dummyConfig.max_size);
       random_height_scale = generate_random_float_in_range(dummyConfig.min_size, dummyConfig.max_size);

       rect2 = rect2.scaled(random_width_scale, random_height_scale)
           .shifted(x_shift, y_shift)
           .global_cropped(0, 0, this.natural_width, this.natural_height);
       this.neppyFiOnce(rect2, dummyConfig);
   }

   neppyFiOnce(random_rect, dummyConfig) {

       const diagonal = Math.sqrt(random_rect.w * random_rect.w + random_rect.h * random_rect.h);

       // let shadow_blur_size = 30;

       const fill_gradient = this.fillctx.createRadialGradient(random_rect.x, random_rect.y, 0, random_rect.x, random_rect.y, diagonal);
       for (const gradientConfigurationElement of dummyConfig.gradient_configuration) {
           fill_gradient.addColorStop(gradientConfigurationElement.offset, gradientConfigurationElement.color);
       }

       const edge_gradient_top = this.fillctx.createLinearGradient(random_rect.x, random_rect.y, random_rect.x2, random_rect.y);
       edge_gradient_top.addColorStop(0.90, "#FFFFFFCF");
       edge_gradient_top.addColorStop(0.98, "#FFFFFF00");

       const edge_gradient_left = this.fillctx.createLinearGradient(random_rect.x, random_rect.y, random_rect.x, random_rect.y2);
       edge_gradient_left.addColorStop(0.90, "#FFFFFFCF");
       edge_gradient_left.addColorStop(0.98, "#FFFFFF00");


       //if (dummyConfig.gradient_configuration.length === 1) {
           // this.fillctx.shadowBlur = shadow_blur_size;
           // this.prepareMask(random_rect.stretched(shadow_blur_size*2, shadow_blur_size*2), bar_configuration);
       //}

       // const old_blur = this.fillctx.shadowBlur;
       // const old_blur_color = this.fillctx.shadowColor;
       this.fillctx.fillStyle = fill_gradient;
       // if (dummyConfig.gradient_configuration.length === 1) {
           // this.fillctx.shadowBlur = shadow_blur_size;
           // this.fillctx.shadowColor = "black";
       // }
       this.fillctx.fillRect(random_rect.x, random_rect.y, random_rect.w, random_rect.h);
       // this.fillctx.shadowBlur = old_blur;
       // this.fillctx.shadowColor = old_blur_color;

       if (dummyConfig.is_border_enabled) {
           this.fillctx.lineWidth = 2;
           this.fillctx.beginPath();
           this.fillctx.strokeStyle = edge_gradient_top;
           this.fillctx.moveTo(random_rect.x + 1, random_rect.y + 1);
           this.fillctx.lineTo(random_rect.x2 - 1, random_rect.y + 1);
           this.fillctx.stroke();

           this.fillctx.beginPath();
           this.fillctx.strokeStyle = edge_gradient_left;
           this.fillctx.moveTo(random_rect.x + 1, random_rect.y2 - 1);
           this.fillctx.lineTo(random_rect.x + 1, random_rect.y + 1);
           this.fillctx.stroke();
       }
   }

   fettyfiArea(image_data, rect, klass, preset_config, cache) {
       rect = rect.scaled(preset_config.scale, preset_config.scale);

       let effect_range = Math.ceil(0.3 * Math.min(rect.h, rect.w) * preset_config.chromatic_aberration_shift_intensity);
       let effect_bounds = rect.stretched(effect_range, effect_range);

       rect = rect.global_cropped(0, 0, this.natural_width, this.natural_height);
       effect_bounds = effect_bounds.global_cropped(0, 0, this.natural_width, this.natural_height);

       let mul = 1;
       if (klass.erotic) {
           mul = 5;
       }
       mul = Math.ceil(mul * preset_config.chromatic_aberration_color_intensity);

       const source_canvas = document.createElement("canvas");
       source_canvas.width = effect_bounds.w;
       source_canvas.height = effect_bounds.h;
       let source_ctx = source_canvas.getContext("2d");
       let sx = effect_bounds.x;
       let sy = effect_bounds.y;
       let sw = effect_bounds.w;
       let sh = effect_bounds.h;
       let dx = 0;
       let dy = 0;
       let dw = effect_bounds.w;
       let dh = effect_bounds.h;
       source_ctx.putImageData(this.clipctx.getImageData(sx, sy, sw, sh), dx, dy);
       //We can prob. save some ms here as this.auxctx.getImageData(sx, sy, sw, sh) (one line above) does already request all the pixels we need
       //so this is kinda redundant what we are doing here and just creates an offset where the array starts to contain pixels...
       //Also: We should do this once for the entire image and not once for each box, this takes 10ms or more to create... so we can save 100ms or more on many images
       const source_data_obj = source_ctx.getImageData(0, 0, source_canvas.width, source_canvas.height);

       let config;

       // Round 1:
       // Green, Horizontal, 50%, default bar & shift sizes
       config = new GlitchConfig();
       config.g_channel_intensity = 0.75 * mul;
       // config.set_channel_intensities(0.05 * mul);
       config.shift_horizontal = preset_config.chromatic_aberration_horizontal_shift;
       config.h_bar_size_range = [Math.floor(rect.h / 20), Math.ceil(rect.h / 10)];
       config.shift_intensity_range = [Math.round(effect_range / 4), effect_range];


       let t1_canvas = document.createElement("canvas");
       t1_canvas.width = effect_bounds.w;
       t1_canvas.height = effect_bounds.h;
       let t1_ctx = t1_canvas.getContext("2d");
       t1_ctx.putImageData(source_ctx.getImageData(dx, dy, dw, dh), dx, dy);


       let mask1 = this.neo_shift(source_ctx, source_data_obj, t1_ctx, t1_canvas, rect, effect_bounds, config);

       // Round 2:
       // Red, Vertical, 200%, default bar & shift sizes.
       config = new GlitchConfig();
       config.r_channel_intensity = 0.85 * (1 + mul);
       // config.set_channel_intensities(0.6 * mul);
       config.shift_vertical = preset_config.chromatic_aberration_vertical_shift;
       config.v_bar_size_range = [Math.floor(rect.w / 20), Math.ceil(rect.w / 10)];
       config.shift_intensity_range = [Math.round(effect_range / 12), Math.round(effect_range / 2)];

       let t2_canvas = document.createElement("canvas");
       t2_canvas.width = effect_bounds.w;
       t2_canvas.height = effect_bounds.h;
       let t2_ctx = t2_canvas.getContext("2d");
       t2_ctx.putImageData(t1_ctx.getImageData(dx, dy, dw, dh), dx, dy);

       /* Todo: enabling these lines will double the duration again, can we somehow improve this? */
       const t1_data_obj = t1_ctx.getImageData(0, 0, t1_canvas.width, t1_canvas.height);
       let mask2 = this.neo_shift(t1_ctx, t1_data_obj, t2_ctx, t2_canvas, rect, effect_bounds, config);
       let mask3 = mask1.union(mask2);
       // let mask3 = mask1;

       let area = effect_bounds.area();
       let logSize = Math.log(area);
       // let sizeScalar = Math.round(logSize);

       const t2_ctx_data = t2_ctx.getImageData(0, 0, t2_canvas.width, t2_canvas.height).data;
       for (let x in mask3.container) {
           if (!mask3.container.hasOwnProperty(x)) continue;
           // x is sourced from the eb_x which is an x from the origin of the eb_rectangle
           x = parseInt(x);
           for (let y in mask3.container[x]) {
               if (!mask3.container[x].hasOwnProperty(y)) continue;
               y = parseInt(y);
               const coordinate_mask = mask3.container[x][y];
               /*
               const fg_pixel = ctx.getImageData(x, y, 1, 1).data;
               const fr = Math.round(fg_pixel[0]);
               const fg = Math.round(fg_pixel[1]);
               const fb = Math.round(fg_pixel[2]);
               */
               const index = (x + y * t2_canvas.width) * 4;
               const fr = t2_ctx_data[index];
               const fg = t2_ctx_data[index + 1];
               const fb = t2_ctx_data[index + 2];
               //const alpha = ctx_data[index+3];
               //Many calculations happen twice here and on top of that we convert colors to different representations
               //just for them to be converted back again as the method just wants the color in this format as the parameter
               //Info: The loop entire itself takes 500ms (without code in it), while these 3 code lines "only" consume 300ms though
               // const fg_color = argb_to_int([0xff, fr, fg, fb]);
               const fg_color = argb_to_int([0xff, fr, fg, fb]) & coordinate_mask;
               const bg_color = getIntColor(image_data, effect_bounds.x + x, effect_bounds.y + y) & ~coordinate_mask;
               setIntColor1(image_data, effect_bounds.x + x, effect_bounds.y + y, fg_color | bg_color);
           }
       }
   }

   glitchAreas(results, cache) {
       let [chromatic_aberration_results, multiple_panels_results] = _.partition(results, (res) => res.preset_config.type === 0);
       if (chromatic_aberration_results.length) {
           let image_data = this.outctx.getImageData(0, 0, this.natural_width, this.natural_height);
           for (const res of chromatic_aberration_results) {
               let { rect, klass, preset_config } = res;
               this.fettyfiArea(image_data, rect, klass, preset_config, cache);
           }
           this.outctx.putImageData(image_data, 0, 0);
       }
       if (multiple_panels_results.length) {
           this.fillctx.reset();

           for (const res of multiple_panels_results) {
               let { rect, klass, preset_config } = res;
               this.neppyfiArea(klass, rect, preset_config, cache);
           }
           this.finishFill(new Rectangle(0, 0, this.natural_width, this.natural_height), cache);
       }
   }

   clipAreasWithShape(results, cache) {
       let results_by_preset_ident = Object.groupBy(results, (res) => res.preset_ident);
       for (let preset_ident in results_by_preset_ident) {
           let results = results_by_preset_ident[preset_ident];
           let preset_config = results[0].preset_config;

           if (preset_config.shape === 4) {
               this.handleLineShapes(results, preset_config, cache, (rect) => {
                   this.fillctx.globalCompositeOperation = "source-atop";
                   this.fillctx.drawImage(this.clip, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);

                   this.outctx.drawImage(this.fill, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
               });
           } else {
               this.handleShapes(results, preset_config, cache, (prefeather_rect, rect) => {
                   this.fillctx.globalCompositeOperation = "source-atop";
                   this.fillctx.drawImage(this.clip, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
               }, (rect) => {
                   this.outctx.drawImage(this.fill, rect.x, rect.y, rect.w, rect.h, rect.x, rect.y, rect.w, rect.h);
               }); 
           }
       }
   }

   // TODO: ditch only mask variable
   applyEffects(censor_type, results, cache) {
       this.effects_applied += 1;
       switch (censor_type) {
           case effects.PIXEL.index:
               this.pixelateAreas(results, cache);
               break;
           case effects.BLUR.index:
               this.blurAreas(results, cache);
               break;
           case effects.BAR.index:
               this.barOnAreas(results, cache);
               break;
           case effects.BOX.index:
               this.boxOnAreas(results, cache);
               break;
           case effects.TRIANGLE.index:
               this.triangulateAreas(results, cache);
               break;
           case effects.GLITCH.index:
                   this.glitchAreas(results, cache);
               break;
           case effects.STICKER.index:
               this.stickerOnAreas(results, cache);
               break;
           case effects.SOBEL.index:
               this.sobelAreas(results, cache);
               break;
           case effects.SPLATTER.index:
               this.splatterAreas(results, cache);
               break;
           case effects.NONE.index:
               break;
       }
   }

   applyNormalEffects(censor_type, normal_results, cache) {
       this.effects_applied += 1;
       switch (censor_type) {
           case effects.PIXEL.index:
               this.clipAreasWithShape(normal_results, cache);
               break;
           case effects.BLUR.index:
               this.clipAreasWithShape(normal_results, cache);
               break;
           case effects.BAR.index:
               this.clipAreasWithShape(normal_results, cache);
               break;
           case effects.BOX.index:
               this.clipAreasWithBox(normal_results, cache);
               break;
           case effects.TRIANGLE.index:
               this.clipAreasWithShape(normal_results, cache);
               break;
           case effects.GLITCH.index:
               // Not valid
               break;
           case effects.STICKER.index:
               this.clipAreasWithSticker(normal_results, cache);
               break;
           case effects.SOBEL.index:
               this.clipAreasWithShape(normal_results, cache);
               break;
           case effects.SPLATTER.index:
               this.clipAreasWithSplatter(normal_results, cache);
               break;
           case effects.NONE.index:
               break;
       }
   }

   applyReverseEffect(censor_type, preset_config, preset_ident, cache) {
       this.outctx.save();
       switch (censor_type) {
           case effects.PIXEL.index:
               this.pixelateCanvas(preset_config, cache);
               break;
           case effects.BLUR.index:
               this.blurCanvas(preset_config, cache);
               break;
           case effects.BAR.index:
               this.barOnCanvas(preset_config, cache);
               break;
           case effects.TRIANGLE.index:
               this.triangulateCanvas(preset_config, cache);
               break;
           case effects.SOBEL.index:
               this.sobelCanvasPrewitt(preset_config, cache);
               break;
           case effects.BOX.index:
               this.boxOnCanvas(preset_config, cache);
               break;
           case effects.GLITCH.index:
               // Not valid
               break;
           case effects.STICKER.index:
               this.stickerOnCanvas(preset_config, preset_ident, cache);
               break;
           case effects.SPLATTER.index:
               this.splatterCanvas(preset_config, cache);
               break;
           case effects.NONE.index:
               break;
       }
       this.outctx.restore();
   }

   before(cache) {
       if (this.word_wall_configuration.enabled) {
           this.ww = document.createElement("canvas");
           this.ww.width = this.natural_width;
           this.ww.height = this.natural_height;
           this.wwctx = this.ww.getContext("2d");
           this.prepareTextMask(this.wwctx, cache);
       }

       if (cache.mixed_mode || cache.reverse_mode) {
           this.clip = document.createElement('canvas');
           this.clip.width = this.natural_width;
           this.clip.height = this.natural_height;
           this.clipctx = this.clip.getContext('2d');
       } else {
           this.clip = this.out;
           this.clipctx = this.outctx;
       }
   }

   beforeLayer(cache) {
       if (cache.mixed_mode || cache.reverse_mode) {
           this.clipctx.drawImage(this.out, 0, 0, this.natural_width, this.natural_height);
       }
   }

   neo_shift(source_context, source_data, target_context, target_canvas, rect, effect_bounds, config) {
       // TODO implement internal H/V distinction.
       let shift_range = config.shift_intensity_range;
       const hShifts = this.calculateShifts(rect.h, config.h_bar_size_range, shift_range);
       const vShifts = this.calculateShifts(rect.w, config.v_bar_size_range, shift_range);

       // TODO per block
       const mask = new Mask();

       // const rect_data_obj = source_context.getImageData(0, 0, rect.w, rect.h);
       // const rect_data = rect_data_obj.data;
       const image_data_obj = source_context.getImageData(0, 0, effect_bounds.w, effect_bounds.h);
       const image_data = image_data_obj.data;

       let i = 0;
       for (let rect_y = 0; rect_y < rect.h; rect_y++) {
           const hShift = hShifts[rect_y];
           if (hShift === undefined) {
               // console.warn("Aborting early as the remaining pixels (" + (rect.h - y) + ") is smaller than the minimum allowed bar height.");
               break;
           }
           const eb_y = rect_y + (rect.y - effect_bounds.y);

           for (let rect_x = 0; rect_x < rect.w; rect_x++) {

               const vShift = vShifts[rect_x];
               if (vShift === undefined) {
                   // console.warn("Aborting early as the remaining pixels (" + (rect.w - x) + ") is smaller than the minimum allowed bar width.");
                   break;
               }
               const eb_x = rect_x + (rect.x - effect_bounds.x);

               // const a = source.getAlphaComponent(x, y);
               //const source_pixel = source_context.getImageData(eb_x, eb_y, 1, 1).data;
               const index = (eb_x + eb_y * source_data.width) * 4;
               const red = source_data.data[index];
               const green = source_data.data[index + 1];
               const blue = source_data.data[index + 2];
               const alpha = source_data.data[index + 3];
               //const source_pixel = rect_data[];
               const sr = Math.round(red * config.r_channel_intensity);
               const sg = Math.round(green * config.g_channel_intensity);
               const sb = Math.round(blue * config.b_channel_intensity);
               const tx = eb_x + hShift;
               const ty = eb_y + vShift;

               // if (!effect_bounds.coordinateInside(tx, ty)) continue;
               /*
               let target_pixel;
               try {
                   target_pixel = source_context.getImageData(tx, ty, 1, 1).data;
               } catch (e) {
                   console.error(rect_x, rect_y, eb_x, eb_y, hShift, vShift, tx, ty, rect);
                   return null;
               }

               const tr = Math.round(target_pixel[0] * (1 - config.r_channel_intensity));
               const tg = Math.round(target_pixel[1] * (1 - config.g_channel_intensity));
               const tb = Math.round(target_pixel[2] * (1 - config.b_channel_intensity));
               const ta = alpha;
               */
               let t_index = (tx + ty * source_data.width) * 4;
               const tr = Math.round(source_data.data[t_index] * (1 - config.r_channel_intensity));
               const tg = Math.round(source_data.data[t_index + 1] * (1 - config.g_channel_intensity));
               const tb = Math.round(source_data.data[t_index + 2] * (1 - config.b_channel_intensity));
               // const tr = 0;
               // const tg = 0;
               // const tb = 0;
               const ta = alpha;

               if (config.r_channel_intensity > 0) mask.setRed(tx, ty);
               if (config.g_channel_intensity > 0) mask.setGreen(tx, ty);
               if (config.b_channel_intensity > 0) mask.setBlue(tx, ty);
               /*
               const image_data = target_context.createImageData(1, 1);
               const id_bytes = image_data.data;
               id_bytes[0] = sr + tr;
               id_bytes[1] = sg + tg;
               id_bytes[2] = sb + tb;
               id_bytes[3] = ta;
               target_context.putImageData(image_data, tx, ty);
               */

               t_index = (tx + ty * image_data_obj.width) * 4;
               image_data_obj.data[t_index] = sr + tr;
               image_data_obj.data[t_index + 1] = sg + tg;
               image_data_obj.data[t_index + 2] = sb + tb;
               image_data_obj.data[t_index + 3] = ta;

           }
       }
       target_context.putImageData(image_data_obj, 0, 0);
       return mask;
   }

   calculateShifts(dimension, bar_range, shift_range) {
       if (bar_range[1] === 0 && bar_range[0] === 0) {
           return new Array(dimension).fill(0);
       }

       let bar_size = 0;
       let shift_intensity = 0;
       const indexed_shifts = [];
       for (let i = 0; i < dimension; i++) {
           if (bar_size > 0) {
               bar_size--;
           } else {
               if (dimension - i < bar_range[1]) {
                   // console.warn("Breaking loop as " + dimension + " - " + i + " < " + bar_range[1] + ".");
                   break;
               }
               bar_size = generate_random_in_range(bar_range[0], bar_range[1]);
               shift_intensity = generate_shift_intensity(shift_range[0], shift_range[1], shift_intensity);
           }
           indexed_shifts[i] = shift_intensity;
       }
       return indexed_shifts;
   }
}

function int_to_argb(int_color) {
   const a_channel = (int_color >> 24) & 0xFF;
   const r_channel = (int_color >> 16) & 0xFF;
   const g_channel = (int_color >> 8) & 0xFF;
   const b_channel = int_color & 0xFF;
   return [a_channel, r_channel, g_channel, b_channel];
}

function argb_to_int(channels) {
   const a_channel = (channels[0] & 0xFF) << 24;
   const r_channel = (channels[1] & 0xFF) << 16;
   const g_channel = (channels[2] & 0xFF) << 8;
   const b_channel = (channels[3] & 0xFF);
   return a_channel | r_channel | g_channel | b_channel;
}

function generate_random_in_range(min, max) {
   const random_value = Math.random() * (max - min);
   return Math.floor(min + random_value);
}

function generate_random_float_in_range(min, max) {
   return min + Math.random() * (max - min);
}

function generate_shift_intensity(min, max, previous) {
   const r_base = min + Math.round(Math.random() * (max - min));
   const r_sign = Math.round(Math.random()) === 0 ? -1 : 1;
   const random_delta = r_base * r_sign;

   let candidate_shift = previous + random_delta;
   if (candidate_shift < -max || max < candidate_shift) {
       candidate_shift = previous - random_delta;
   }
   return candidate_shift;
}

function generateBaseBlurImproved(canvas, blurring, grayscale) {
   let blurredCanvas = document.createElement("canvas");
   let pxExt = blurring * 2;
   let pxExt2 = pxExt * 2;
   blurredCanvas.width = canvas.width + pxExt2;
   blurredCanvas.height = canvas.height + pxExt2;
   const ctx = blurredCanvas.getContext("2d");
   if (grayscale) {
       ctx.filter = "grayscale(1)";
   }
   ctx.save();
   //Upper left corner
   ctx.translate(pxExt, pxExt);
   ctx.scale(-1, -1);
   ctx.drawImage(canvas, 0, 0, pxExt, pxExt, 0, 0, pxExt, pxExt);
   ctx.restore();
   ctx.save();
   //Top
   ctx.translate(0, pxExt);
   ctx.scale(1, -1);
   ctx.drawImage(canvas, 0, 0, canvas.width, pxExt, pxExt, 0, canvas.width, pxExt);
   ctx.restore();
   ctx.save();
   //Upper right corner
   ctx.translate(blurredCanvas.width, pxExt);
   ctx.scale(-1, -1);
   ctx.drawImage(canvas, canvas.width - pxExt, 0, pxExt, pxExt, 0, 0, pxExt, pxExt);
   ctx.restore();
   ctx.save();
   //Left
   ctx.translate(pxExt, 0);
   ctx.scale(-1, 1);
   ctx.drawImage(canvas, 0, 0, pxExt, canvas.height, 0, pxExt, pxExt, canvas.height,
   );
   ctx.restore();
   ctx.save();
   //Right
   ctx.translate(blurredCanvas.width, 0);
   ctx.scale(-1, 1);
   ctx.drawImage(canvas, canvas.width - pxExt, 0, pxExt, canvas.height, 0, pxExt, pxExt, canvas.height);
   ctx.restore();
   ctx.save();
   //Bottom left corner
   ctx.translate(pxExt, blurredCanvas.height);
   ctx.scale(-1, -1);
   ctx.drawImage(canvas, 0, canvas.height - pxExt, pxExt, pxExt, 0, 0, pxExt, pxExt);
   ctx.restore();
   ctx.save();
   //Bottom
   ctx.translate(0, blurredCanvas.height);
   ctx.scale(1, -1);
   ctx.drawImage(canvas, 0, canvas.height - pxExt, canvas.width, pxExt, pxExt, 0, canvas.width, pxExt);
   ctx.restore();
   ctx.save();
   //Bottom right corner
   ctx.translate(blurredCanvas.width, blurredCanvas.height);
   ctx.scale(-1, -1);
   ctx.drawImage(canvas, canvas.width - pxExt, canvas.height - pxExt, pxExt, pxExt, 0, 0, pxExt, pxExt);
   ctx.restore();
   ctx.save();
   //Center
   ctx.drawImage(canvas, pxExt, pxExt, canvas.width, canvas.height);
   ctx.filter = "blur(" + blurring + "px)";
   ctx.drawImage(blurredCanvas, 0, 0, blurredCanvas.width, blurredCanvas.height);
   let blurred = document.createElement("canvas");
   blurred.width = canvas.width;
   blurred.height = canvas.height;
   const ctx_blurred = blurred.getContext("2d");
   ctx_blurred.drawImage(blurredCanvas, pxExt, pxExt, canvas.width, canvas.height, 0, 0, canvas.width, canvas.height);

   return blurred;
}

function mosaicPixel(ctx, rect, pixel_size, spacing, clip, clipctx, grayscale, color_precision, accurate_sampling, operation = "source-over") {
   ctx.globalCompositeOperation = operation;
   // TODO: reuse this image data
   let pixel_arr = clipctx.getImageData(0, 0, clip.width, clip.height).data;
   let start_x = previousMultiple(rect.x, pixel_size + spacing);
   let start_y = previousMultiple(rect.y, pixel_size + spacing);
   for (let x = start_x; x < rect.x + rect.w; x += pixel_size + spacing) {
       for (let y = start_y; y < rect.y + rect.h; y += pixel_size + spacing) {
           ctx.fillStyle = calculatePixelBoxColor(clip.width, x + pixel_size / 2, y + pixel_size / 2, pixel_size, pixel_arr, grayscale, color_precision, accurate_sampling);
           ctx.fillRect(x, y, pixel_size, pixel_size);
       }
   }
}

function toColor(num, alpha = false) {
   num >>>= 0;
   let b = num & 0xFF, g = (num & 0xFF00) >>> 8, r = (num & 0xFF0000) >>> 16, a = ((num & 0xFF000000) >>> 24) / 255;
   if (!alpha) {
       a = 1;
   }
   return "rgba(" + [r, g, b, a].join(",") + ")";
}
function calculateHexagonSide(area) {
   return Math.sqrt(area / (3 * Math.sqrt(3) / 2));
}

function hexagonalPixel(ctx, rect, pixel_size, spacing, clip, clipctx, grayscale, color_precision, accurate_sampling, operation = "source-over") {
   let a = 2 * Math.PI / 6;
   let r = calculateHexagonSide(pixel_size * pixel_size) + spacing;

   ctx.strokeStyle = "#000";
   ctx.lineWidth = spacing;
   function paintPixel(ctx, x, y, r, a, pixel_arr) {
       ctx.beginPath();
       for (let i = 0; i < 6; i++) {
           ctx.lineTo(x + (r + 0.5) * Math.cos(a * i), y + (r + 0.5) * Math.sin(a * i));
       }
       ctx.globalCompositeOperation = operation;
       ctx.fillStyle = calculatePixelBoxColor(clip.width, x, y, r * 2, pixel_arr, grayscale, color_precision, accurate_sampling);

       ctx.closePath();
       ctx.fill();
       if (spacing) {
           ctx.globalCompositeOperation = "destination-out";
           ctx.stroke();
       }
   }

   let pixel_arr = clipctx.getImageData(0, 0, clip.width, clip.height).data;

   let x_step = r * (1 + Math.cos(a));
   let y_step = 2 * (r * Math.sin(a));
   let start_x_i = Math.floor(rect.x / x_step);
   let start_x = start_x_i * x_step;
   let start_y = previousMultiple(rect.y, y_step) + 0.5 * ((-1) ** start_x_i * r * Math.sin(a));
   for (let y = start_y; y < rect.y + rect.h + r; y += y_step) {
       for (let x = start_x, j = 0; x < rect.x + rect.w + r; x += x_step) {
           let yc = y + 0.5 * ((-1) ** j++ * r * Math.sin(a));
           paintPixel(ctx, x, yc, r, a, pixel_arr);
       }
   }
}

function glitchedPixel(ctx, rect, pixel_size, spacing, clip, clipctx, grayscale, color_precision, accurate_sampling, operation = "source-over") {
   let a = Math.ceil(2 * Math.PI / 6);
   let r = calculateHexagonSide(pixel_size * pixel_size);

   ctx.globalCompositeOperation = operation;
   function paintPixel(ctx, x, y, r, a, pixel_arr) {
       ctx.beginPath();
       for (let i = 0; i < 6; i++) {
           ctx.lineTo(x + r * Math.cos(a * i), y + r * Math.sin(a * i));
       }
       ctx.fillStyle = calculatePixelBoxColor(clip.width, x, y, r, pixel_arr, grayscale, color_precision, accurate_sampling);

       ctx.closePath();
       ctx.fill();
   }

   let pixel_arr = clipctx.getImageData(0, 0, clip.width, clip.height).data;

   let x_step = r * (1 + Math.cos(a));
   let y_step = 2 * (r * Math.sin(a));
   let start_x_i = Math.floor(rect.x / x_step + spacing);
   let start_x = start_x_i * x_step + spacing;
   let start_y = previousMultiple(rect.y, y_step + spacing) + 0.5 * ((-1) ** start_x_i * r * Math.sin(a));
   for (let y = start_y; y < rect.y + rect.h + r; y += y_step) {
       y += spacing;
       for (let x = start_x, j = 0; x < rect.x + rect.w + r; x += x_step) {
           x += spacing;
           let yc = y + 0.5 * ((-1) ** j++ * r * Math.sin(a));
           paintPixel(ctx, x, yc, r, a, pixel_arr);
       }
   }
}

function calculateSmoothing(rect, config) {
   let smoothing_strength = Math.ceil(Math.min(rect.w, rect.h) / 100 * config.feathering);
   return smoothing_strength;
}

function resizeFile(imageFile, size = 80) {
   let resolver = () => {
   };
   let reader = new FileReader();
   reader.onload = function (e) {
       let img = document.createElement("img");
       img.onload = function (event) {
           let canvas = document.createElement("canvas");
           canvas.width = size;
           canvas.height = size;
           let ctx = canvas.getContext("2d");
           ctx.drawImage(img, 0, 0, size, size);
           let imageData = ctx.getImageData(0, 0, size, size);
           resolver(imageData);
       };
       img.src = e.target.result;
   };
   reader.readAsDataURL(imageFile);
   return new Promise((resolve, reject) => {
       resolver = resolve;
   });
}

function mergeImageData(imageDataArray) {

   let newImageData = imageDataArray[0];

   for (let j = 0; j < imageDataArray.length; j++) { // iterate through the imageDataArray

       for (let i = 0, bytes = imageDataArray[j].data.length; i < bytes; i += 4) { // iterate through image bytes

           let index = (imageDataArray[j].data[i + 3] === 0 ? 0 : j);

           newImageData.data[i] = imageDataArray[index].data[i];
           newImageData.data[i + 1] = imageDataArray[index].data[i + 1];
           newImageData.data[i + 2] = imageDataArray[index].data[i + 2];
           newImageData.data[i + 3] = imageDataArray[index].data[i + 3];
       }
   }
   return newImageData;
}

/**
* Draws a rounded rectangle using the current state of the canvas.
* If you omit the last three params, it will draw a rectangle
* outline with a 5 pixel border radius
* @param {CanvasRenderingContext2D} ctx
* @param {Number} x The top left x coordinate
* @param {Number} y The top left y coordinate
* @param {Number} width The width of the rectangle
* @param {Number} height The height of the rectangle
* @param {Number} [radius = 5] The corner radius; It can also be an object
*                 to specify different radii for corners
* @param {Number} [radius.tl = 0] Top left
* @param {Number} [radius.tr = 0] Top right
* @param {Number} [radius.br = 0] Bottom right
* @param {Number} [radius.bl = 0] Bottom left
* @param {Boolean} [fill = false] Whether to fill the rectangle.
* @param {Boolean} [stroke = true] Whether to stroke the rectangle.
*/
function roundRect(ctx, x, y, width, height, radius = 5, fill = false, stroke = false) {
   if (typeof radius === "number") {
       radius = {tl: radius, tr: radius, br: radius, bl: radius};
   } else {
       radius = {...{tl: 0, tr: 0, br: 0, bl: 0}, ...radius};
   }
   ctx.beginPath();
   ctx.moveTo(x + radius.tl, y);
   ctx.lineTo(x + width - radius.tr, y);
   ctx.quadraticCurveTo(x + width, y, x + width, y + radius.tr);
   ctx.lineTo(x + width, y + height - radius.br);
   ctx.quadraticCurveTo(x + width, y + height, x + width - radius.br, y + height);
   ctx.lineTo(x + radius.bl, y + height);
   ctx.quadraticCurveTo(x, y + height, x, y + height - radius.bl);
   ctx.lineTo(x, y + radius.tl);
   ctx.quadraticCurveTo(x, y, x + radius.tl, y);
   ctx.closePath();
   if (fill) {
       ctx.fill();
   }
   if (stroke) {
       ctx.stroke();
   }
}

function toPrevMultiple(n, multiple) {
   return Math.floor(n / multiple) * multiple;
}

function toNextMultiple(n, multiple) {
   return Math.ceil(n / multiple) * multiple;
}

function toClosestMultiple(n, multiple) {
   return Math.round(n / multiple) * multiple;
}

/**
* Calculates the color of a pixel box based on the provided parameters.
* @param {number} width - The width of the image.
* @param {number} x - The unrounded x-coordinate of the center of the pixel box.
* @param {number} y - The unrounded y-coordinate of the center of the pixel box.
* @param {number} px_size - The size of the pixel box.
* @param {number[]} px_arr - The array of pixel values.
* @param {boolean} grayscale
* @param {number} color_precision
* @param {boolean} accurate_sampling
* @returns {string} The color of the pixel box in the format "rgba(r, g, b, a)".
*/
function calculatePixelBoxColor(width, x, y, px_size, px_arr, grayscale, color_precision, accurate_sampling) {
   if (grayscale) {
       let avg_color, a;
       if (accurate_sampling) {
           let height = px_arr.length / (width * 4);
           avg_color = 0, a = 0;
           let px_count = 0;
           for (let i = Math.max(0, Math.round(x - px_size / 2)); i < Math.min(width, x + px_size / 2); i++) {
               for (let j = Math.max(0, Math.round(y - px_size / 2)); j < Math.min(height, y + px_size / 2); j++) {
                   let index = (i + (j * width)) * 4;
                   px_count++;
                   avg_color += ((px_arr[index] + px_arr[index + 1] + px_arr[index + 2]) / 3 - avg_color) / px_count;
                   a += (px_arr[index + 3] - a) / px_count;
               }
           }
       } else {
           let index = (Math.round(x) + (Math.round(y) * width)) * 4;
           avg_color = (px_arr[index] + px_arr[index + 1] + px_arr[index + 2]) / 3;
           a = px_arr[index + 3];
       }

       return "rgba(" 
           + Math.round(toClosestMultiple(avg_color, color_precision)) + "," 
           + Math.round(toClosestMultiple(avg_color, color_precision)) + "," 
           + Math.round(toClosestMultiple(avg_color, color_precision)) + "," 
           + a + ")";
   } else {
       let r, g, b, a;
       if (accurate_sampling) {
           let height = px_arr.length / (width * 4);
           r = 0, g = 0, b = 0, a = 0;
           let px_count = 0;
           for (let i = Math.max(0, Math.round(x - px_size / 2)); i < Math.min(width, x + px_size / 2); i++) {
               for (let j = Math.max(0, Math.round(y - px_size / 2)); j < Math.min(height, y + px_size / 2); j++) {
                   let index = (i + (j * width)) * 4;
                   px_count++;
                   r += (px_arr[index] - r) / px_count;
                   g += (px_arr[index + 1] - g) / px_count;
                   b += (px_arr[index + 2] - b) / px_count;
                   a += (px_arr[index + 3] - a) / px_count;
               }
           }
       } else {
           let index = (Math.round(x) + (Math.round(y) * width)) * 4;
           r = px_arr[index];
           g = px_arr[index + 1];
           b = px_arr[index + 2];
           a = px_arr[index + 3];
       }

       return "rgba(" 
           + Math.round(toClosestMultiple(r, color_precision)) + "," 
           + Math.round(toClosestMultiple(g, color_precision)) + "," 
           + Math.round(toClosestMultiple(b, color_precision)) + "," 
           + a + ")";
   }
}

function handleOOMImage(data, blob, file, filter, stream, callback) {
   let urlCreator = window.URL || window.webkitURL;
   let imageUrl = urlCreator.createObjectURL(blob);
   let image = new Image();
   image.onload = function () {
       let naturalWidth = this.naturalWidth;
       let naturalHeight = this.naturalHeight;
       let out = document.createElement("canvas");
       out.width = naturalWidth;
       out.height = naturalHeight;
       let numDate = new Date(data.date);
       let classes = data.classes;
       let outctx = out.getContext("2d");
       let px = 40;
       let ratio = px / 1000;
       let fontSize = naturalWidth * ratio;
       let blur_px = only_once_mode_configuration.mode_configuration[5].blur * Math.max(naturalWidth, naturalHeight);
       let promises = [];
       if (data.date > Date.now()) {
           outctx.drawImage(image, 0, 0, naturalWidth, naturalHeight);
       } else {
           switch (only_once_mode_configuration.mode) {
               case 1: // Slight Transparency
                   outctx.drawImage(image, 0, 0, naturalWidth, naturalHeight);
                   let opc = 1 - only_once_mode_configuration.mode_configuration[1].transparency;
                   outctx.fillStyle = "rgba(0,0,0," + opc + ")";
                   outctx.fillRect(0, 0, naturalWidth, naturalHeight);
                   break;
               case 2: // Borderless
                   outctx.drawImage(image, 0, 0, naturalWidth, naturalHeight);
                   let distance = only_once_mode_configuration.mode_configuration[2].distance;
                   let radius = only_once_mode_configuration.mode_configuration[2].radius;
                   let border = naturalWidth * distance;
                   //ctx.fillRect(border,border,naturalWidth-border*2,naturalHeight-border*2);
                   roundRect(outctx, border, border, naturalWidth - border * 2, naturalHeight - border * 2, radius, true, false);
                   break;
               case 3: // Preview
                   outctx.fillStyle = "rgba(0,0,0)";
                   outctx.fillRect(0, 0, naturalWidth, naturalHeight);
                   let scaled_width = naturalWidth * 0.1;
                   let scaled_height = naturalHeight * 0.1;
                   let x = naturalWidth / 2 - scaled_width / 2;
                   let y = naturalHeight / 2 - scaled_height - fontSize;
                   outctx.save();
                   blur_px = only_once_mode_configuration.mode_configuration[3].blur * Math.max(scaled_width, scaled_height);
                   outctx.filter = "blur(" + blur_px + "px)";
                   roundedImage(outctx, x, y, scaled_width, scaled_height, 10);
                   outctx.clip();
                   outctx.drawImage(image, x, y, scaled_width, scaled_height);
                   outctx.restore();
                   break;
               case 4: // Line Pattern
                   outctx.drawImage(image, 0, 0, naturalWidth, naturalHeight);
                   let strength = only_once_mode_configuration.mode_configuration[4].strength;
                   let numberOfStripes = Math.min(naturalWidth, naturalHeight) * strength;
                   //let color1 = "rgb(0,0,0,1)",color2="rgba(255,255,255,0.55)";
                   let color1 = only_once_mode_configuration.mode_configuration[4].color_1;
                   let color2 = only_once_mode_configuration.mode_configuration[4].color_2;
                   for (let i = 0; i < numberOfStripes * 2; i++) {
                       let thickness = (naturalWidth / numberOfStripes);
                       outctx.beginPath();
                       outctx.strokeStyle = i % 2 ? color1 : color2;
                       outctx.lineWidth = thickness;
                       outctx.lineCap = "round";
                       outctx.moveTo(i * thickness + thickness / 2 - naturalWidth, 0);
                       outctx.lineTo(0 + i * thickness + thickness / 2, naturalHeight);
                       outctx.stroke();
                   }
                   for (let i = 0; i < numberOfStripes * 2; i++) {
                       let thickness = (naturalWidth / numberOfStripes);
                       outctx.beginPath();
                       outctx.strokeStyle = i % 2 ? color1 : color2;
                       outctx.lineWidth = thickness;
                       outctx.lineCap = "round";
                       outctx.moveTo(i * thickness + thickness / 2, 0);
                       outctx.lineTo(0 + i * thickness + thickness / 2 - naturalWidth, naturalHeight);
                       outctx.stroke();
                   }
                   break;
               case 5: // Strong Blur
                   let grayscale = only_once_mode_configuration.mode_configuration[5].grayscale;
                   outctx.save();
                   outctx.filter = "blur(" + blur_px + "px)" + (grayscale ? " grayscale(100%)" : "");
                   outctx.drawImage(image, 0, 0, naturalWidth, naturalHeight);
                   outctx.restore();
                   break;
               case 6: // Box Tease
                   outctx.drawImage(image, 0, 0, naturalWidth, naturalHeight);
                   let p = new Promise((resolve, reject) => {
                       let frame = outctx.getImageData(0, 0, naturalWidth, naturalHeight);
                       let payload = {
                           ai_request: true, frame: frame, width: naturalWidth, height: naturalHeight,
                       };
                       handleFrameMessage(payload, null, function (ai_data) {
                           let config = bundleConfiguration();
                           let orig_box_config = config.box_configuration;
                           config.box_configuration = structuredClone(orig_box_config);
                           config.box_configuration.presets.forEach((preset) => {
                               if (preset == null) return;
                               preset.config.fill = 0;
                               preset.config.border = 0;
                           })
                           config.reverse_mode_configuration = {...config.reverse_mode_configuration, enabled: true, mode: 0 };

                           let cache = {};
                           let labels = 
                               !only_once_mode_configuration.mode_configuration[6].allow_faces 
                               ? all_labels 
                               : all_labels.filter((label) => label !== klasses.FACE_FEMALE.key && label !== klasses.FACE_MALE.key);

                           let processedRequest = new ProcessedCensorRequest(imageUrl, file, naturalWidth, naturalHeight, "png", ai_data.boxes, ai_data.scores, ai_data.classes, ai_data.valid_detections, "box", labels);
                           paintCensorRequest(out, outctx, processedRequest, config, cache);
                           resolve();
                       });
                   });
                   promises.push(p);
                   break;
               default: // Full Block
                   outctx.fillStyle = "rgba(0,0,0)";
                   outctx.fillRect(0, 0, naturalWidth, naturalHeight);
           }
       }

       Promise.allSettled(promises).then(() => {
           if (data.date > Date.now()) {
               paintOOMFutureTimestamp(data, out, outctx);
           } else {
               outctx.font = fontSize + "px Segoe UI";
               outctx.fillStyle = "white";
               outctx.textBaseline = "middle";
               outctx.textAlign = "center";
               let sblur = 2;
               let slinewidth = 2;
               let text = only_once_mode_configuration.message;
               let m_text = outctx.measureText(text);
               outctx.shadowColor = "black";
               outctx.shadowBlur = sblur;
               outctx.lineWidth = slinewidth;
               outctx.strokeText(text, naturalWidth / 2, naturalHeight / 2 + (m_text.actualBoundingBoxAscent - m_text.actualBoundingBoxDescent) / 2);
               outctx.shadowBlur = 0;
               outctx.fillText(text, naturalWidth / 2, naturalHeight / 2 + (m_text.actualBoundingBoxAscent - m_text.actualBoundingBoxDescent) / 2);
               let y = naturalHeight / 2;
               if (only_once_mode_configuration.date_time_format != null && only_once_mode_configuration.date_time_format !== "") {
                   let text_date = "" + numDate.toLocaleString();
                   if (only_once_mode_configuration.date_time_format !== "local") {
                       text_date = "" + numDate.toLocaleString(only_once_mode_configuration.date_time_format);
                   }
                   let m_date = outctx.measureText(text_date);
                   y += (m_date.actualBoundingBoxAscent - m_date.actualBoundingBoxDescent) / 2 + fontSize;
                   outctx.shadowColor = "black";
                   outctx.shadowBlur = sblur;
                   outctx.lineWidth = slinewidth;
                   outctx.strokeText(text_date, naturalWidth / 2, y);
                   outctx.shadowBlur = 0;
                   outctx.fillText(text_date, naturalWidth / 2, y);
               }
               if (classes && only_once_mode_configuration.display_classes) {
                   for (const [index, value] of Object.entries(classes)) {
                       let r = Object.values(klasses).find((e) => e.index == index);
                       let text_label = r.name + ": " + value;
                       let m_text_label = outctx.measureText(text_label);
                       y += (m_text_label.actualBoundingBoxAscent - m_text_label.actualBoundingBoxDescent) / 2 + fontSize;
                       outctx.shadowColor = "black";
                       outctx.shadowBlur = sblur;
                       outctx.lineWidth = slinewidth;
                       outctx.strokeText(text_label, naturalWidth / 2, y);
                       outctx.shadowBlur = 0;
                       outctx.fillText(text_label, naturalWidth / 2, y);
                   }
               }
               out.toBlob(function (blob) {
                   if (filter != null) {
                       blob.arrayBuffer().then(buffer => {
                           filter.write(buffer);
                           filter.disconnect();
                           if (callback) {
                               callback(blob);
                           }
                       });
                   } else if (callback) {
                       callback(blob);
                   }
               });
               URL.revokeObjectURL(imageUrl);
           }
       });
   };
   image.src = imageUrl;
}

function paintOOMFutureTimestamp(data, canvas, ctx) {
   ctx.save();
   let text = "Permanently blocked in ";
   let naturalWidth = canvas.width;
   let px = 14;
   let ratio = px / 1000;
   let fontSize = naturalWidth * ratio;
   let sblur = 2;
   let slinewidth = 2;
   ctx.font = fontSize + "px Segoe UI";
   ctx.fillStyle = "white";
   ctx.shadowBlur = sblur;
   ctx.lineWidth = slinewidth;
   ctx.shadowColor = "black";
   const diffTime = Math.abs(Date.now() - data.date);
   text += msToTime(diffTime);
   let m_text = ctx.measureText(text);
   ctx.fillText(text, naturalWidth - m_text.width - px, px * 1.2 + (m_text.actualBoundingBoxAscent - m_text.actualBoundingBoxDescent) / 2);
   ctx.shadowBlur = 0;
   ctx.fillText(text, naturalWidth - m_text.width - px, px * 1.2 + (m_text.actualBoundingBoxAscent - m_text.actualBoundingBoxDescent) / 2);
   ctx.restore();

   /*
   let text = "This image will be permanently blocked:";
   let naturalWidth = image.naturalWidth;
   let naturalHeight = image.naturalHeight;
   let px = 40;
   let ratio = px / 1000;
   let fontSize = naturalWidth * ratio;
   let sblur = 2;
   let slinewidth = 2;
   let numDate= new Date(data.date);
   ctx.save();
   ctx.font = fontSize+"px Segoe UI";
   ctx.fillStyle = "white";
   ctx.textBaseline = 'middle';
   ctx.textAlign = 'center';
   let m_text = ctx.measureText(text);
   ctx.shadowColor="black";
   ctx.shadowBlur=sblur;
   ctx.lineWidth=slinewidth;
   ctx.strokeText(text, naturalWidth / 2, naturalHeight / 2 + (m_text.actualBoundingBoxAscent - m_text.actualBoundingBoxDescent) / 2);
   ctx.shadowBlur=0;
   ctx.fillText(text, naturalWidth / 2, naturalHeight / 2 + (m_text.actualBoundingBoxAscent - m_text.actualBoundingBoxDescent) / 2);
   let y = naturalHeight / 2;
   if(only_once_mode_configuration.date_time_format != null){
       let text_date = ""+numDate.toLocaleString();
       if(only_once_mode_configuration.date_time_format !== 'local'){
           text_date = ""+numDate.toLocaleString(only_once_mode_configuration.date_time_format);
       }
       let m_date = ctx.measureText(text_date);
       y += (m_date.actualBoundingBoxAscent - m_date.actualBoundingBoxDescent) / 2 + fontSize;
       ctx.shadowColor="black";
       ctx.shadowBlur=sblur;
       ctx.lineWidth=slinewidth;
       ctx.strokeText(text_date, naturalWidth / 2, y);
       ctx.shadowBlur=0;
       ctx.fillText(text_date, naturalWidth / 2, y);
   }
   ctx.restore();
   */
}